import contextlib
import copyreg
from weakref import WeakSet

from dill._dill import _is_builtin_module
from spacy.tokens import Doc, Span, Token
from spacy.vocab import Vocab

_pickle_memo = WeakSet()


def _reduce_doc(doc):
    seen = doc in _pickle_memo

    if not seen:
        # uid = uuid.uuid4()
        _pickle_memo.add(doc)
        # -> [uid, need to encode body]
        return (_add_back_user_data_and_hooks, (doc, doc.user_data, doc.user_hooks))

    # uid, need_to_encode_body = _pickle_memo[doc]
    # if not need_to_encode_body:
    #     return (_rebuild_doc, (uid,))

    array_head = Doc._get_array_attrs()
    strings = set()
    for token in doc:
        strings.add(token.tag_)
        strings.add(token.lemma_)
        strings.add(str(token.morph))
        strings.add(token.dep_)
        strings.add(token.ent_type_)
        strings.add(token.ent_kb_id_)
        strings.add(token.ent_id_)
        strings.add(token.norm_)
    for group in doc.spans.values():
        for span in group:
            strings.add(span.label_)
            if span.kb_id in span.doc.vocab.strings:
                strings.add(span.kb_id_)
            if span.id in span.doc.vocab.strings:
                strings.add(span.id_)
    data = {
        "vocab": doc.vocab,
        "text": doc.text,
        "array_head": array_head,
        "array_body": doc.to_array(array_head),
        "sentiment": doc.sentiment,
        "tensor": doc.tensor,
        "cats": doc.cats,
        "spans": doc.spans.to_bytes(),
        "strings": list(strings),
        "has_unknown_spaces": doc.has_unknown_spaces,
    }
    _pickle_memo.remove(doc)
    return (_rebuild_doc, (data,))


def _rebuild_doc(data):
    doc = Doc(Vocab())
    doc.from_dict(data)
    return doc


def _add_back_user_data_and_hooks(doc, user_data, user_hooks):
    doc.user_data = user_data
    doc.user_hooks = user_hooks
    return doc


def _reduce_span(span):
    data = {
        "start": span.start,
        "end": span.end,
        "label": span.label_,
        "kb_id": span.kb_id_,
        "id": span.id_,
        "doc": span.doc,
    }
    return (_rebuild_span, (data,))


def _rebuild_span(data):
    return Span(
        doc=data["doc"],
        start=data["start"],
        end=data["end"],
        label=data["label"],
        kb_id=data["kb_id"],
        span_id=data["id"],
    )


def _reduce_token(token):
    return (
        _rebuild_token,
        (
            token.doc,
            token.i,
        ),
    )


def _rebuild_token(doc, i):
    """Import the token contents from a dictionary.

    doc (Doc): The parent document.
    i (int): The index of the token in the document.
    RETURNS (Token): The deserialized `Token`.
    """
    return doc[i]


def custom_is_builtin_module(module):
    if not hasattr(module, "__file__"):
        return True
    if module.__file__ is None:
        return False
    # Hack to avoid copying the full module dict
    return True


@contextlib.contextmanager
def pickler_dont_save_module_dict():
    old_is_builtin_module_code = _is_builtin_module.__code__
    old_is_builtin_module_defaults = _is_builtin_module.__defaults__

    _is_builtin_module.__code__ = custom_is_builtin_module.__code__
    _is_builtin_module.__defaults__ = custom_is_builtin_module.__defaults__

    yield

    _is_builtin_module.__code__ = old_is_builtin_module_code
    _is_builtin_module.__defaults__ = old_is_builtin_module_defaults


copyreg.pickle(Doc, _reduce_doc)
copyreg.pickle(Span, _reduce_span)
copyreg.pickle(Token, _reduce_token)



from itertools import islice
from pathlib import Path

import pytest
from typing_extensions import Literal

import edsnlp


@pytest.mark.parametrize("num_cpu_workers", [0, 2])
@pytest.mark.parametrize("shuffle", ["dataset"])
def test_read_shuffle_loop(
    num_cpu_workers: int,
    shuffle: Literal["dataset", "fragment"],
):
    input_file = (
        Path(__file__).parent.parent.resolve() / "training" / "rhapsodie_sample.conllu"
    )
    notes = edsnlp.data.read_conll(
        input_file,
        shuffle=shuffle,
        seed=42,
        loop=True,
    ).set_processing(num_cpu_workers=num_cpu_workers)
    notes = list(islice(notes, 6))
    assert len(notes) == 6
    # 32	ce	ce	PRON	_	Gender=Masc|Number=Sing|Person=3|PronType=Dem	30	obl:arg	_	_  # noqa: E501
    word_attrs = {
        "text": "ce",
        "lemma_": "ce",
        "pos_": "PRON",
        "dep_": "obl:arg",
        "morph": "Gender=Masc|Number=Sing|Person=3|PronType=Dem",
        "head": "profité",
    }
    word = notes[0][31]
    for attr, val in word_attrs.items():
        assert str(getattr(word, attr)) == val



import importlib
import subprocess
import sys
import tarfile
import zipfile

import pytest

import edsnlp
from edsnlp.package import package


def test_blank_package(nlp, tmp_path):
    # Missing metadata makes poetry fail due to missing author / description
    if not isinstance(nlp, edsnlp.Pipeline):
        pytest.skip("Only running for edsnlp.Pipeline")

    package(
        pipeline=nlp,
        root_dir=tmp_path,
        name="test-model-fail",
        metadata={},
        project_type="poetry",
    )

    nlp.package(
        root_dir=tmp_path,
        name="test-model",
        metadata={
            "description": "A test model",
            "authors": "Test Author <test.author@mail.com>",
        },
        project_type="poetry",
        distributions=["wheel"],
    )
    assert (tmp_path / "dist").is_dir()
    assert (tmp_path / "dist" / "test_model-0.1.0-py3-none-any.whl").is_file()
    assert not (tmp_path / "dist" / "test_model-0.1.0.tar.gz").is_file()


@pytest.mark.parametrize("package_name", ["my-test-model", None])
@pytest.mark.parametrize("manager", ["poetry", "setuptools"])
def test_package_with_files(nlp, tmp_path, package_name, manager):
    if not isinstance(nlp, edsnlp.Pipeline):
        pytest.skip("Only running for edsnlp.Pipeline")

    nlp.to_disk(tmp_path / "model", exclude=set())

    ((tmp_path / "test_model").mkdir(parents=True))
    (tmp_path / "test_model" / "__init__.py").write_text('print("Hello World!")\n')
    (tmp_path / "test_model" / "empty_folder").mkdir()
    (tmp_path / "README.md").write_text(
        """\
<!-- INSERT -->
# Test Model
"""
    )
    if manager == "poetry":
        (tmp_path / "pyproject.toml").write_text(
            """\
[build-system]
requires = ["poetry-core>=1.0.0"]
build-backend = "poetry.core.masonry.api"

[tool.poetry]
name = "test-model"
version = "0.0.0"
description = "A test model"
authors = ["Test Author <test.author@mail.com>"]
readme = "README.md"

[tool.poetry.dependencies]
python = ">=3.7"
build = "*"  # sample light package to install
"""
        )
    elif manager == "setuptools":
        (tmp_path / "pyproject.toml").write_text(
            """\
[build-system]
requires = ["setuptools>=42", "wheel"]
build-backend = "setuptools.build_meta"

[project]
name = "test-model"
version = "0.0.0"
description = "A test model"
authors = [
    {name = "Test Author", email = "test.author@mail.com"}
]
readme = "README.md"
requires-python = ">=3.7"

dependencies = [
    "build"
]
"""
        )
    package(
        name=package_name,
        pipeline=tmp_path / "model",
        root_dir=tmp_path,
        check_dependencies=False,
        version="0.1.0",
        distributions=None,
        metadata={
            "description": "A new description",
            "authors": "Test Author <test.author@mail.com>",
        },
        readme_replacements={
            "<!-- INSERT -->": "Replaced !",
        },
    )

    module_name = "test_model" if package_name is None else "my_test_model"

    assert (tmp_path / "dist").is_dir()
    assert (tmp_path / "dist" / f"{module_name}-0.1.0.tar.gz").is_file()
    assert (tmp_path / "dist" / f"{module_name}-0.1.0-py3-none-any.whl").is_file()
    assert (tmp_path / "pyproject.toml").is_file()

    with zipfile.ZipFile(
        tmp_path / "dist" / f"{module_name}-0.1.0-py3-none-any.whl"
    ) as zf:
        # check files
        assert set(zf.namelist()) == {
            f"{module_name}-0.1.0.dist-info/METADATA",
            f"{module_name}-0.1.0.dist-info/RECORD",
            f"{module_name}-0.1.0.dist-info/WHEEL",
            f"{module_name}/__init__.py",
            f"{module_name}/artifacts/config.cfg",
            f"{module_name}/artifacts/meta.json",
            f"{module_name}/artifacts/tokenizer",
            "test_model/__init__.py",
        }
        # check description
        with zf.open(f"{module_name}-0.1.0.dist-info/METADATA") as f:
            assert b"A new description" in f.read()

    with tarfile.open(tmp_path / "dist" / f"{module_name}-0.1.0.tar.gz") as tf:
        # check files
        assert set(tf.getnames()) == {
            f"{module_name}-0.1.0/PKG-INFO",
            f"{module_name}-0.1.0/README.md",
            f"{module_name}-0.1.0/artifacts/config.cfg",
            f"{module_name}-0.1.0/artifacts/meta.json",
            f"{module_name}-0.1.0/artifacts/tokenizer",
            f"{module_name}-0.1.0/{module_name}/__init__.py",
            f"{module_name}-0.1.0/pyproject.toml",
            f"{module_name}-0.1.0/test_model/__init__.py",
        }
        # check description
        with tf.extractfile(f"{module_name}-0.1.0/PKG-INFO") as f:
            assert b"A new description" in f.read()

        with tf.extractfile(f"{module_name}-0.1.0/README.md") as f:
            assert b"Replaced !" in f.read()

    # pip install the whl file
    (tmp_path / "site-packages").mkdir(exist_ok=True)
    subprocess.check_output(
        [
            sys.executable,
            "-m",
            "pip",
            "install",
            "-vvv",
            "--target",
            str(tmp_path / "site-packages"),
            str(tmp_path / "dist" / f"{module_name}-0.1.0-py3-none-any.whl"),
        ],
        stderr=subprocess.STDOUT,
    )

    site_packages = tmp_path / "site-packages"
    sys.path.insert(0, str(site_packages))
    # check site-package files
    files = {str(f.relative_to(site_packages)) for f in set(site_packages.rglob("*"))}
    assert files >= {
        f"{module_name}/artifacts",
        f"{module_name}/artifacts/config.cfg",
        f"{module_name}/artifacts/meta.json",
        f"{module_name}/artifacts/tokenizer",
    }

    module = importlib.import_module(module_name)

    with open(module.__file__) as f:
        assert f.read() == (
            ('print("Hello World!")\n' if package_name is None else "")
            + """
# -----------------------------------------
# This section was autogenerated by edsnlp
# -----------------------------------------

import edsnlp
from pathlib import Path
from typing import Optional, Dict, Any

__version__ = '0.1.0'

def load(
    overrides: Optional[Dict[str, Any]] = None,
) -> edsnlp.Pipeline:
    path_outside = Path(__file__).parent / "../artifacts"
    path_inside = Path(__file__).parent / "artifacts"
    path = path_inside if path_inside.exists() else path_outside
    model = edsnlp.load(path, overrides=overrides)
    return model
"""
        )
    module.load()
    edsnlp.load(module_name)



from __future__ import annotations

from collections import defaultdict
from typing import Any, Dict, Iterable, List, Optional, Set

from spacy.tokens import Doc, Span
from typing_extensions import Literal

from edsnlp.core.pipeline import Pipeline
from edsnlp.pipes.trainable.embeddings.typing import (
    WordEmbeddingComponent,
)
from edsnlp.pipes.trainable.ner_crf.ner_crf import NERBatchOutput, TrainableNerCrf
from edsnlp.utils.filter import align_spans, filter_spans
from edsnlp.utils.span_getters import (
    SpanGetterArg,
    SpanSetterArg,
    get_spans,
)
from edsnlp.utils.typing import AsList


class TrainableExtractiveQA(TrainableNerCrf):
    """
    The `eds.extractive_qa` component is a trainable extractive question answering
    component. This can be seen as a Named Entity Recognition (NER) component where the
    types of entities predicted by the model are not pre-defined during the training
    but are provided as prompts (i.e., questions) at inference time.

    The `eds.extractive_qa` shares a lot of similarities with the `eds.ner_crf`
    component, and therefore most of the arguments are the same.

    !!! note "Extractive vs Abstractive Question Answering"

        Extractive Question Answering differs from Abstractive Question Answering in
        that the answer is extracted from the text, rather than generated (à la
        ChatGPT) from scratch. To normalize the answers, you can use the
        `eds.span_linker` component in `synonym` mode and search for the closest
        `synonym` in a predefined list.

    Examples
    --------

    ```python
    import edsnlp, edsnlp.pipes as eds

    nlp = edsnlp.blank("eds")
    nlp.add_pipe(
        eds.extractive_qa(
            embedding=eds.transformer(
                model="prajjwal1/bert-tiny",
                window=128,
                stride=96,
            ),
            mode="joint",
            target_span_getter="ner-gold",
            span_setter="ents",
            questions={
                "disease": "What disease does the patient have?",
                "drug": "What drug is the patient taking?",
            },  # (1)!
        ),
        name="qa",
    )
    ```

    To train the model, refer to the [Training](/tutorials/make-a-training-script)
    tutorial.

    Once the model is trained, you can use the questions attribute (next section) on the
    document you run the model on, or you can change the global questions attribute:

    ```python
    nlp.pipes.qa.questions = {
        "disease": "When did the patient get sick?",
    }
    ```

    # Dynamic Questions

    You can also provide

    ```{ .python .no-check }
    eds.extractive_qa(..., questions_attribute="questions")
    ```

    to get the questions dynamically from an attribute on the Doc or Span objects
    (e.g., `doc._.questions`). This is useful when you want to have different questions
    depending on the document.

    To provide questions from a dataframe, you can use the following code:

    ```{ .python .no-check }
    dataframe = pd.DataFrame({"questions": ..., "note_text": ..., "note_id": ...})
    stream = edsnlp.data.from_pandas(
        dataframe,
        converter="omop",
        doc_attributes={"questions": "questions"},
    )
    stream.map_pipeline(nlp)
    stream.set_processing(backend="multiprocessing")
    out = stream.to_pandas(converters="ents")
    ```


    Parameters
    ----------
    name : str
        Name of the component
    embedding : WordEmbeddingComponent
        The word embedding component
    questions : Dict[str, AsList[str]]
        The questions to ask, as a mapping between the entity type and the list of
        questions to ask for this entity type (or single string if only one question).
    questions_attribute : Optional[str]
        The attribute to use to get the questions dynamically from the Doc or Span
        objects (as returned by the `context_getter` argument). If None, the questions
        will be fixed and only taken from the `questions` argument.
    context_getter : Optional[SpanGetterArg]
        What context to use when computing the span embeddings (defaults to the whole
        document). For example `{"section": "conclusion"}` to only extract the
        entities from the conclusion.
    target_span_getter : SpanGetterArg
        Method to call to get the gold spans from a document, for scoring or training.
        By default, takes all entities in `doc.ents`, but we recommend you specify
        a given span group name instead.
    span_setter : Optional[SpanSetterArg]
        The span setter to use to set the predicted spans on the Doc object. If None,
        the component will infer the span setter from the target_span_getter config.
    infer_span_setter : Optional[bool]
        Whether to complete the span setter from the target_span_getter config.
        False by default, unless the span_setter is None.
    mode : Literal["independent", "joint", "marginal"]
        The CRF mode to use : independent, joint or marginal
    window : int
        The window size to use for the CRF. If 0, will use the whole document, at
        the cost of a longer computation time. If 1, this is equivalent to assuming
        that the tags are independent and will the component be faster, but with
        degraded performance. Empirically, we found that a window size of 10 or 20
        works well.
    stride : Optional[int]
        The stride to use for the CRF windows. Defaults to `window // 2`.
    """

    def __init__(
        self,
        nlp: Optional[Pipeline] = None,
        name: Optional[str] = "extractive_qa",
        *,
        embedding: WordEmbeddingComponent,
        questions: Dict[str, AsList[str]] = {},
        questions_attribute: str = "questions",
        context_getter: Optional[SpanGetterArg] = None,
        target_span_getter: Optional[SpanGetterArg] = None,
        span_setter: Optional[SpanSetterArg] = None,
        infer_span_setter: Optional[bool] = None,
        mode: Literal["independent", "joint", "marginal"] = "joint",
        window: int = 40,
        stride: Optional[int] = None,
    ):
        self.questions_attribute: Optional[str] = questions_attribute
        self.questions = questions
        super().__init__(
            nlp=nlp,
            name=name,
            embedding=embedding,
            context_getter=context_getter,
            span_setter=span_setter,
            target_span_getter=target_span_getter,
            mode=mode,
            window=window,
            stride=stride,
            infer_span_setter=infer_span_setter,
        )
        self.update_labels(["answer"])
        self.labels_to_idx = defaultdict(lambda: 0)

    def set_extensions(self):
        super().set_extensions()
        if self.questions_attribute:
            if not Doc.has_extension(self.questions_attribute):
                Doc.set_extension(self.questions_attribute, default=None)
            if not Span.has_extension(self.questions_attribute):
                Span.set_extension(self.questions_attribute, default=None)

    def post_init(self, docs: Iterable[Doc], exclude: Set[str]):
        pass

    @property
    def cfg(self):
        cfg = dict(super().cfg)
        cfg.pop("labels")
        return cfg

    def preprocess(self, doc, **kwargs):
        contexts = (
            list(get_spans(doc, self.context_getter))
            if self.context_getter
            else [doc[:]]
        )
        prompt_contexts_and_labels = sorted(
            {
                (prompt, label, context)
                for context in contexts
                for label, questions in (
                    *self.questions.items(),
                    *(getattr(doc._, self.questions_attribute) or {}).items(),
                    *(
                        (getattr(context._, self.questions_attribute) or {}).items()
                        if context is not doc
                        else ()
                    ),
                )
                for prompt in questions
            }
        )
        questions = [x[0] for x in prompt_contexts_and_labels]
        labels = [x[1] for x in prompt_contexts_and_labels]
        ctxs = [x[2] for x in prompt_contexts_and_labels]
        lengths = [len(ctx) for ctx in ctxs]
        return {
            "lengths": lengths,
            "$labels": labels,
            "$contexts": ctxs,
            "embedding": self.embedding.preprocess(
                doc,
                contexts=ctxs,
                prompts=questions,
                **kwargs,
            ),
            "stats": {"ner_words": sum(lengths)},
        }

    def preprocess_supervised(self, doc, **kwargs):
        prep = self.preprocess(doc, **kwargs)
        contexts = prep["$contexts"]
        labels = prep["$labels"]
        tags = []

        for context, label, target_ents in zip(
            contexts,
            labels,
            align_spans(
                list(get_spans(doc, self.target_span_getter)),
                contexts,
            ),
        ):
            span_tags = [[0] * len(self.labels) for _ in range(len(context))]
            start = context.start
            target_ents = [ent for ent in target_ents if ent.label_ == label]

            # TODO: move this to the LinearChainCRF class
            for ent in filter_spans(target_ents):
                label_idx = self.labels_to_idx[ent.label_]
                if ent.start == ent.end - 1:
                    span_tags[ent.start - start][label_idx] = 4
                else:
                    span_tags[ent.start - start][label_idx] = 2
                    span_tags[ent.end - 1 - start][label_idx] = 3
                    for i in range(ent.start + 1 - start, ent.end - 1 - start):
                        span_tags[i][label_idx] = 1
            tags.append(span_tags)

        return {
            **prep,
            "targets": tags,
        }

    def postprocess(
        self,
        docs: List[Doc],
        results: NERBatchOutput,
        inputs: List[Dict[str, Any]],
    ):
        spans: Dict[Doc, list[Span]] = defaultdict(list)
        contexts = [ctx for sample in inputs for ctx in sample["$contexts"]]
        labels = [label for sample in inputs for label in sample["$labels"]]
        tags = results["tags"].cpu()
        for context_idx, _, start, end in self.crf.tags_to_spans(tags).tolist():
            span = contexts[context_idx][start:end]
            label = labels[context_idx]
            span.label_ = label
            spans[span.doc].append(span)
        for doc in docs:
            self.set_spans(doc, spans.get(doc, []))
        return docs



import copy
import re
from difflib import SequenceMatcher
from typing import List, Tuple

import numpy as np
import regex
from typing_extensions import NotRequired, TypedDict


class DeltaCollection(object):
    """
    Copied from
    https://github.com/percevalw/nlstruct/blob/master/nlstruct/data_utils.py#L13
    """

    def __init__(self, begins, ends, deltas):
        self.begins = np.asarray(begins, dtype=int)
        self.ends = np.asarray(ends, dtype=int)
        self.deltas = np.asarray(deltas, dtype=int)

    def apply(self, positions, side="left"):
        positions = np.asarray(positions, dtype=int)
        to_add = (
            (positions.reshape(-1, 1) >= self.ends.reshape(1, -1)) * self.deltas
        ).sum(axis=1)
        between = np.logical_and(
            self.begins.reshape(1, -1) < positions.reshape(-1, 1),
            positions.reshape(-1, 1) < self.ends.reshape(1, -1),
        )
        between_mask = between.any(axis=1)
        between = between[between_mask]
        if len(between) > 0:
            between_i = between.argmax(axis=1)
            if side == "right":
                to_add[between_mask] += (
                    self.ends[between_i]
                    - positions[between_mask]
                    + self.deltas[between_i]
                )
            elif side == "left":
                to_add[between_mask] += self.begins[between_i] - positions[between_mask]
        return positions + to_add

    def unapply(self, positions, side="left"):
        positions = np.asarray(positions, dtype=int)
        begins = self.apply(self.begins, side="left")
        ends = self.apply(self.ends, side="right")
        to_remove = -(
            (positions.reshape(-1, 1) >= ends.reshape(1, -1)) * self.deltas
        ).sum(axis=1)
        between = np.logical_and(
            begins.reshape(1, -1) < positions.reshape(-1, 1),
            positions.reshape(-1, 1) < ends.reshape(1, -1),
        )
        between_mask = between.any(axis=1)
        between = between[between_mask]
        pos = positions + to_remove
        if len(between) > 0:
            between_i = between.argmax(axis=1)
            if side == "right":
                pos[between_mask] = self.ends[between_i]
            elif side == "left":
                pos[between_mask] = self.begins[between_i]
        return pos

    def __add__(self, other):
        if len(self.begins) == 0:
            return other
        if len(other.begins) == 0:
            return self
        begins = self.unapply(other.begins, side="left")
        ends = self.unapply(other.ends, side="right")
        new_begins = np.concatenate([begins, self.begins])
        new_ends = np.concatenate([ends, self.ends])
        new_deltas = np.concatenate([other.deltas, self.deltas])
        sorter = np.lexsort((new_ends, new_begins))
        return DeltaCollection(new_begins[sorter], new_ends[sorter], new_deltas[sorter])


def make_str_from_groups(replacement, groups):
    for i, group in enumerate(groups):
        group = group or ""
        replacement = replacement.replace(f"\\{i + 1}", group).replace(
            f"\\g<{i + 1}>", group
        )
    return replacement


def regex_sub_with_spans(pattern, replacement, text):
    needed_groups = [
        int(next(j for j in i if j))
        for i in regex.findall(r"\\([0-9]+)|\\g<([0-9]+)>", replacement)
    ]
    begins = []
    ends = []
    deltas = []
    for match in reversed(list(regex.finditer(pattern, text, flags=regex.DOTALL))):
        middle = make_str_from_groups(
            replacement, [match.group(i) for i in needed_groups]
        )
        start = match.start()
        end = match.end()
        text = text[:start] + middle + text[end:]
        begins.append(start)
        ends.append(end)
        deltas.append(len(middle) - end + start)
    return text, DeltaCollection(begins, ends, deltas)


def regex_multisub_with_spans(
    patterns, replacements, text, deltas=None, return_deltas=False
):
    deltas = DeltaCollection([], [], []) if deltas is None and return_deltas else None
    for pattern, replacement in zip(patterns, replacements):
        if return_deltas:
            text, new_deltas = regex_sub_with_spans(pattern, replacement, text)
            deltas += new_deltas
        else:
            text = regex.sub(pattern, replacement, text)
    return text, deltas


def similar(a, b):
    return SequenceMatcher(None, a, b).ratio()


def get_left_context(fragment, text, n=20):
    return re.sub(r"\s+", " ", text[: fragment["begin"]])[-n:]


def get_right_context(fragment, text, n=20):
    return re.sub(r"\s+", " ", text[fragment["end"] :])[:n]


def remove_spaces(doc, shift=False):
    doc = copy.deepcopy(doc)
    doc["text"], deltas = regex_multisub_with_spans(
        [r"\s+"], [" "], doc["text"], return_deltas=True
    )
    if shift:
        doc = shift_spans(doc, deltas)
    return doc, deltas


def shift_spans(doc, deltas, reverse=False):
    fragments = [f for e in doc["entities"] for f in e["fragments"]]
    lefts = [f["begin"] for f in fragments]
    rights = [f["end"] for f in fragments]
    if not reverse:
        new_lefts = deltas.apply(lefts, side="left")
        new_rights = deltas.apply(rights, side="right")
    else:
        new_lefts = deltas.unapply(lefts, side="left")
        new_rights = deltas.unapply(rights, side="right")
    for f, b, e in zip(fragments, new_lefts.tolist(), new_rights.tolist()):
        f["begin"] = b
        f["end"] = e
    return doc


Fragment = TypedDict("Fragment", {"begin": int, "end": int})

AnnotatedText = TypedDict(
    "AnnotatedText",
    {
        "doc_id": NotRequired[str],
        "text": str,
        "entities": List[
            TypedDict(
                "Entity",
                {"label": str, "fragments": List[Fragment]},
            )
        ],
    },
)

Result = TypedDict(
    "Result",
    {
        "total_count": int,
        "missing": List[Fragment],
        "good": List[Fragment],
        "unsure": List[Fragment],
        "missing_count": int,
        "good_count": int,
        "unsure_count": int,
        "doc": AnnotatedText,
    },
)


def align(
    old: AnnotatedText,
    new: AnnotatedText,
    sim_scheme: List[Tuple[int, float]] = [
        (20, 0.70),
        (50, 0.20),
        (100, 0.15),
        (400, 0.10),
        (1000, 0.05),
    ],
    threshold: float = 1.0,
    do_debug: bool = False,
) -> Result:
    """
    Align entities from two similar, but not identical documents. The entities
    of the old document are aligned to the new document based on string matching
    and context similarity, and annotated to the new document.

    Unlike approaches relying on diffing algorithms, this method can handle
    insertions and deletions and swaps of text blocks.

    This method was developed during our work on
    [edspdf](https://github.com/aphp/edspdf), to transfer annotations
    from dataset annotated on previous versions of documents to
    newer versions of the same documents. Yet, there is significant
    room for improvement, especially in the similarity scoring scheme.


    Parameters
    ----------
    old : AnnotatedText
        The old document with entities to align.
    new : AnnotatedText
        The new document to align entities to.
    sim_scheme : List[Tuple[int, float]]
        A list of (context size, weight) tuples to use for similarity scoring.
        Each tuple defines how many characters of left and right context to consider
        and the weight of that context in the overall similarity score.
    threshold : float
        The similarity score threshold above which a match is considered good.
        It's computed as a weighted sum of left and right similarity scores from
        different context sizes.
    do_debug : bool
        Whether to print debug information.

    Returns
    -------
    Result
    """
    missing = []
    good = []
    unsure = []

    # Normalize spaces from both docs (and keep track of character shifts)
    old, _ = remove_spaces(old, shift=True)
    new_raw_text = new["text"]
    new, deltas = remove_spaces(new)

    if do_debug:
        debug = print
        print("Doc ID:", new.get("doc_id", None))
    else:

        def debug(*args, **kwargs):  # pragma: no cover
            return None

    old_text = old["text"]
    new_text = new["text"]

    # List fragments in old doc
    old_fragments = [
        {"label": e["label"], **f} for e in old["entities"] for f in e["fragments"]
    ]
    new_fragments = []

    # Try and find offsets in the new document for every old fragment
    for fragment in old_fragments:
        begin_is_end = False
        if (fragment["begin"], fragment["end"]) == (0, 0):  # pragma: no cover
            new_fragments.append(fragment)
            continue

        fragment_text = old_text[fragment["begin"] : fragment["end"]]
        original_fragment_text = fragment_text

        if len(fragment_text) <= 2:
            try:
                fragment_text = old_text[fragment["begin"] :].split()[0]
            except IndexError:  # pragma: no cover
                fragment_text = ""

        if len(fragment_text) <= 2:
            fragment_text = old_text[fragment["begin"] : fragment["begin"] + 10]

        virtual_fragment = {
            **fragment,
            "end": fragment["begin"] + len(fragment_text),
        }
        snippets = [
            (
                get_left_context(virtual_fragment, old_text, n),
                get_right_context(virtual_fragment, old_text, n),
            )
            for n, w in sim_scheme
        ]

        candidates = []
        for match in re.finditer(re.escape(fragment_text), new["text"]):
            candidates.append({"begin": match.start(), "end": match.end()})

        if len(original_fragment_text) <= 2 and len(candidates) > 20:
            debug(
                "Too small: {}[{}]{}".format(
                    snippets[0][0], fragment_text, snippets[0][1]
                )
            )
            missing.append(fragment)
            new_fragments.append(None)
            continue

        if len(candidates) == 0:
            begin_is_end = True
            prev_snippets = snippets
            prev_fragment_text = fragment_text
            fragment_text = old_text[fragment["end"] : fragment["end"] + 10]
            virtual_fragment = {
                **fragment,
                "begin": fragment["end"],
                "end": fragment["end"] + 10,
            }
            snippets = [
                (
                    get_left_context(virtual_fragment, old_text, n),
                    get_right_context(virtual_fragment, old_text, n),
                )
                for n, w in sim_scheme
            ]
            candidates = []
            for match in re.finditer(re.escape(fragment_text), new["text"]):
                candidates.append({"begin": match.start(), "end": match.end()})

            if len(original_fragment_text) <= 2 and len(candidates) > 20:
                debug(
                    "Too small: {}[{}]{}".format(
                        snippets[0][0], fragment_text, snippets[0][1]
                    ),
                    prev_fragment_text,
                )
                missing.append(fragment)
                new_fragments.append(None)
                continue

            if len(candidates) == 0:
                debug(
                    "! Missing !: {}[{}]{}".format(
                        prev_snippets[0][0], prev_fragment_text, prev_snippets[0][1]
                    )
                )
                missing.append(fragment)
                new_fragments.append(None)
                continue

        for candidate in candidates:
            new_snippets = [
                (
                    get_left_context(candidate, new_text, n),
                    get_right_context(candidate, new_text, n),
                )
                for n, w in sim_scheme
            ]
            score = sum(
                (similar(old_l, new_l) + similar(old_r, new_r)) * w
                for ((old_l, old_r), (new_l, new_r), (n, w)) in zip(
                    snippets, new_snippets, sim_scheme
                )
            ) / sum(w for n, w in sim_scheme)

            candidate["score"] = score
            candidate["new_snippets"] = new_snippets

        best_score = max((c["score"] for c in candidates))

        if begin_is_end:
            snippets = prev_snippets
            fragment_text = prev_fragment_text

        if best_score > threshold:
            good.append(fragment)
            candidate = max(candidates, key=lambda c: c["score"])
            if begin_is_end:
                new_fragments.append(
                    {
                        **fragment,
                        "begin": candidate["begin"]
                        - (fragment["end"] - fragment["begin"]),
                        "end": candidate["begin"],
                    }
                )
            else:
                new_fragments.append(
                    {
                        **fragment,
                        "begin": candidate["begin"],
                        "end": candidate["begin"] + fragment["end"] - fragment["begin"],
                    }
                )
        else:
            unsure.append(fragment)
            new_fragments.append(None)

            debug("Unsure about ", repr(fragment_text))
            debug(
                "    {}[{}]{}".format(snippets[0][0], fragment_text, snippets[0][1])
                .ljust(50)
                .replace("\n", " "),
                "(ORIGINAL)",
            )
            debug("    ------ ↓ candidates ↓ ------")
            for candidate in candidates:
                new_snippets = candidate["new_snippets"]
                score_str = (
                    str(candidate["score"])
                    if candidate["score"] != best_score
                    else "\033[91;1m{}\n\033[0m".format(str(candidate["score"]))
                )
                debug(
                    "    {}[{}]{} : {}".format(
                        new_snippets[0][0],
                        fragment_text,
                        new_snippets[0][1],
                        score_str,
                    ).replace("\n", " ")
                )
            debug()

    # Reconstruct new document from new fragments
    new_fragments_iter = iter(new_fragments)
    new_doc = {
        **old,
        "text": new_raw_text,
        "entities": [
            {
                **e,
                "fragments": [
                    {"begin": new_f["begin"], "end": new_f["end"]}
                    for old_f, new_f in zip(e["fragments"], new_fragments_iter)
                    if new_f is not None
                ],
            }
            for e in old["entities"]
        ],
    }
    new_doc["entities"] = [
        e for e in new_doc["entities"] if all(f is not None for f in e["fragments"])
    ]
    # shift it since we have offsets in the new document with normalized spaces
    new_doc = shift_spans(new_doc, deltas, reverse=True)

    return {
        "total_count": len(old_fragments),
        "missing": missing,
        "good": good,
        "unsure": unsure,
        "missing_count": len(missing),
        "good_count": len(good),
        "unsure_count": len(unsure),
        "doc": new_doc,
    }



from typing import TYPE_CHECKING, Any, Dict, Generic, List, Optional

from foldedtensor import FoldedTensor
from spacy.tokens import Doc, Span
from typing_extensions import TypedDict

from edsnlp.core.torch_component import BatchInput, TorchComponent  # noqa: F401
from edsnlp.utils.span_getters import SpanGetter

WordEmbeddingBatchOutput = TypedDict(
    "WordEmbeddingBatchOutput",
    {
        "embeddings": FoldedTensor,
    },
)


class WordEmbeddingComponent(
    TorchComponent[WordEmbeddingBatchOutput, BatchInput],
    Generic[BatchInput],
):
    span_getter: Optional[SpanGetter]
    output_size: int

    if TYPE_CHECKING:

        def preprocess(
            self,
            doc: Doc,
            *,
            contexts: Optional[List[Span]],
            **kwargs,
        ) -> Dict[str, Any]:
            ...


class WordContextualizerComponent(
    Generic[BatchInput], WordEmbeddingComponent[BatchInput]
):
    span_getter: Optional[SpanGetter]
    output_size: int
    embedding: WordEmbeddingComponent


SpanEmbeddingBatchOutput = TypedDict(
    "SpanEmbeddingBatchOutput",
    {
        "embeddings": FoldedTensor,
    },
)


class SpanEmbeddingComponent(
    TorchComponent[SpanEmbeddingBatchOutput, BatchInput],
    Generic[BatchInput],
):
    span_getter: Optional[SpanGetter]
    output_size: int

    if TYPE_CHECKING:

        def preprocess(
            self,
            doc: Doc,
            *,
            spans: Optional[List[Span]],
            contexts: Optional[List[Span]],
            pre_aligned: bool = False,
            **kwargs,
        ) -> Dict[str, Any]:
            ...



import random
import re
from typing import Iterable, Optional

from spacy.tokens import Doc, Span

import edsnlp
from edsnlp import Pipeline

EMPTY = object()


def make_shifter(start, end, new_doc):
    cache = {}

    def rec(obj):
        if isinstance(obj, Span):
            if obj in cache:
                return cache[obj]
            if obj.end > start and obj.start < end:
                res = Span(
                    new_doc,
                    max(0, obj.start - start),
                    min(obj.end - start, end - start),
                    obj.label,
                )
            else:
                res = EMPTY
            cache[obj] = res
        elif isinstance(obj, (list, tuple, set)):
            res = type(obj)(
                filter(
                    lambda x: x is not EMPTY,
                    (rec(span) for span in obj),
                )
            )
        elif isinstance(obj, dict):
            res = {}
            for k, v in obj.items():
                new_v = rec(v)
                if new_v is not EMPTY:
                    res[k] = new_v
        else:
            res = obj
        return res

    return rec


def subset_doc(doc: Doc, start: int, end: int) -> Doc:
    """
    Subset a doc given a start and end index.

    Parameters
    ----------
    doc: Doc
        The doc to subset
    start: int
        The start index
    end: int
        The end index

    Returns
    -------
    Doc
    """
    new_doc = doc[start:end].as_doc()

    shifter = make_shifter(start, end, new_doc)

    char_beg = doc[start].idx if start < len(doc) else 0
    char_end = doc[end - 1].idx + len(doc[end - 1].text)
    for k, val in list(doc.user_data.items()):
        new_value = shifter(val)
        if k[0] == "._." and new_value is not EMPTY:
            new_doc.user_data[
                (
                    k[0],
                    k[1],
                    None if k[2] is None else max(0, k[2] - char_beg),
                    None if k[3] is None else min(k[3] - char_beg, char_end - char_beg),
                )
            ] = new_value

    for name, group in doc.spans.items():
        new_doc.spans[name] = shifter(list(group))

    return new_doc


@edsnlp.registry.factory.register("eds.split", spacy_compatible=False)
class Split:
    def __init__(
        self,
        nlp: Optional[Pipeline] = None,
        name: str = "split",
        *,
        max_length: int = 0,
        regex: Optional[str] = "\n{2,}",
        filter_expr: Optional[str] = None,
        randomize: float = 0.0,
    ):
        """
        The `eds.split` component splits a document into multiple documents
        based on a regex pattern or a maximum length.

        !!! warning "Not for pipelines"

            This component is not meant to be used in a pipeline, but rather
            as a preprocessing step when dealing with a stream of documents
            as in the example below.

        Examples
        --------

        ```python
        import edsnlp, edsnlp.pipes as eds

        # Create the stream
        stream = edsnlp.data.from_iterable(
            ["Sentence 1\\n\\nThis is another longer sentence more than 5 words"]
        )

        # Convert texts into docs
        stream = stream.map_pipeline(edsnlp.blank("eds"))

        # Apply the split component
        stream = stream.map(eds.split(max_length=5, regex="\\n{2,}"))

        print(" | ".join(doc.text.strip() for doc in stream))
        # Out: Sentence 1 | This is another longer sentence | more than 5 words
        ```

        Parameters
        ----------
        nlp: Optional[Pipeline]
            The pipeline
        name: str
            The component name
        max_length: int
            The maximum length of the produced documents.
            If 0, the document will not be split based on length.
        regex: Optional[str]
            The regex pattern to split the document on
        filter_expr: Optional[str]
            An optional filter expression to filter the produced documents. The callable
            expects a single `doc` argument, the new Doc, and should return a boolean.
            For example, to filter out documents with less than 5 tokens,
            you can use `filter_expr="len(doc) >= 5"`.
        randomize: float
            The randomization factor to split the documents, to avoid
            producing documents that are all `max_length` tokens long
            (0 means all documents will have the maximum possible length
            while 1 will produce documents with a length varying between
            0 and `max_length` uniformly)
        """
        self.max_length = max_length
        self.regex = re.compile(regex) if regex else None
        self.filter_fn = eval(f"lambda doc:{filter_expr}") if filter_expr else None
        self.randomize = randomize

    def __call__(self, doc: Doc) -> Iterable[Doc]:
        for sub_doc in self.split_doc(doc):
            if sub_doc.text.strip():
                if not self.filter_fn or self.filter_fn(sub_doc):
                    yield sub_doc

    def split_doc(
        self,
        doc: Doc,
    ) -> Iterable[Doc]:
        """
        Split a doc into multiple docs of max_length tokens.

        Parameters
        ----------
        doc: Doc
            The doc to split

        Returns
        -------
        Iterable[Doc]
        """
        max_length = self.max_length

        if max_length <= 0 and self.regex is None:
            yield doc
            return

        start = 0
        end = 0
        # Split doc into segments between the regex matches
        matches = (
            [
                next(
                    m.end(g)
                    for g in range(self.regex.groups + 1)
                    if m.end(g) is not None
                )
                for m in self.regex.finditer(doc.text)
            ]
            if self.regex
            else []
        ) + [len(doc.text)]
        word_ends = doc.to_array("IDX") + doc.to_array("LENGTH")
        segments_end = word_ends.searchsorted([m for m in matches], side="right")

        for end in segments_end:
            # If the sentence adds too many tokens
            if end - start > max_length > 0:
                # But the current buffer too large
                while end - start > max_length:
                    subset_end = (
                        start + int(max_length * (random.random() ** self.randomize))
                        if self.randomize
                        else start + max_length
                    )
                    yield subset_doc(doc, start, subset_end)
                    start = subset_end
                yield subset_doc(doc, start, end)
                start = end

            if end > start:
                yield subset_doc(doc, start, end)
                start = end

        yield subset_doc(doc, start, end)



import os
import re
from pathlib import Path
from typing import Optional, Tuple, Union

import fsspec.implementations.local
import pyarrow.fs
from fsspec import AbstractFileSystem
from fsspec import __version__ as fsspec_version
from fsspec.implementations.arrow import ArrowFSWrapper

FileSystem = Union[AbstractFileSystem, pyarrow.fs.FileSystem]

if fsspec_version < "2023.3.0":
    # Ugly hack to make fsspec's arrow implementation work in python 3.7
    # since arrow requires files to be seekable, and the default fsspec
    # open(..., seekable) parameter is False
    # See https://github.com/fsspec/filesystem_spec/pull/1186
    ArrowFSWrapper._open.__wrapped__.__defaults__ = ("rb", None, True)


def walk_match(
    fs: FileSystem,
    root: str,
    file_pattern: str,
) -> list:
    if fs.isfile(root):
        return [root]
    return [
        os.path.join(dirpath, f)
        for dirpath, dirnames, files in fs.walk(root)
        for f in files
        if re.match(file_pattern, f)
    ]


def normalize_fs_path(
    filesystem: Optional[FileSystem],
    path: Union[str, Path],
) -> Tuple[AbstractFileSystem, str]:
    has_protocol = isinstance(path, str) and "://" in path
    filesystem = (
        ArrowFSWrapper(filesystem)
        if isinstance(filesystem, pyarrow.fs.FileSystem)
        else filesystem
    )

    # We need to detect the fs from the path
    if filesystem is None or has_protocol:
        uri: str = path if has_protocol else f"file://{os.path.abspath(path)}"
        inferred_fs, fs_path = fsspec.core.url_to_fs(uri)
        inferred_fs: fsspec.AbstractFileSystem
        filesystem = filesystem or inferred_fs
        assert inferred_fs.protocol == filesystem.protocol, (
            f"Protocol {inferred_fs.protocol} in path does not match "
            f"filesystem {filesystem.protocol}"
        )
        path = fs_path  # path without protocol

    return (
        ArrowFSWrapper(filesystem)
        if isinstance(filesystem, pyarrow.fs.FileSystem)
        else filesystem
    ), str(path)



import warnings
from datetime import date, datetime

from dateutil.parser import parse as parse_date
from spacy.tokens import Doc

if not Doc.has_extension("note_id"):
    Doc.set_extension("note_id", default=None)


def set_note_datetime(doc, dt):
    try:
        if type(dt) is datetime:
            pass
        elif isinstance(dt, str):
            dt = parse_date(dt)
        elif isinstance(dt, (int, float)):
            dt = datetime.fromtimestamp(dt)
        elif isinstance(dt, date):
            dt = datetime(dt.year, dt.month, dt.day)
        elif dt is None:
            pass
        key = doc._._get_key("note_datetime")
        doc.doc.user_data[key] = dt
        return
    except Exception:
        pass

    warnings.warn(f"Cannot cast {dt} as a note datetime", UserWarning)


def get_note_datetime(doc):
    key = doc._._get_key("note_datetime")
    return doc.user_data.get(key, None)


if not Doc.has_extension("note_datetime"):
    Doc.set_extension(
        "note_datetime",
        getter=get_note_datetime,
        setter=set_note_datetime,
    )

if not Doc.has_extension("birth_datetime"):
    Doc.set_extension("birth_datetime", default=None)



import pytest

try:
    import optuna
except ImportError:
    optuna = None

if optuna is None:
    pytest.skip("optuna not installed", allow_module_level=True)


from edsnlp.tune import update_config


@pytest.fixture
def minimal_config():
    return {
        "train": {
            "layers": None,
        },
        ".lr": {
            "learning_rate": None,
        },
    }


@pytest.fixture
def hyperparameters():
    return {
        "train.layers": {
            "type": "int",
            "low": 2,
            "high": 8,
            "step": 2,
        },
        "'.lr'.learning_rate": {
            "alias": "learning_rate",
            "type": "float",
            "low": 0.001,
            "high": 0.1,
            "log": True,
        },
        "train.batch_size": {
            "alias": "batch_size",
            "type": "categorical",
            "choices": [32, 64, 128],
        },
    }


@pytest.fixture
def hyperparameters_with_invalid_type():
    return {
        "train.optimizer": {
            "type": "string",
            "choices": ["adam", "sgd"],
        }
    }


@pytest.fixture
def hyperparameters_with_invalid_path():
    return {
        "model.layers": {
            "type": "int",
            "low": 2,
            "high": 8,
            "step": 2,
        },
    }


@pytest.fixture
def trial():
    study = optuna.create_study(direction="maximize")
    trial = study.ask()
    return trial


def test_update_config_with_values(minimal_config, hyperparameters):
    values = {"learning_rate": 0.05, "train.layers": 6, "batch_size": 64}
    _, updated_config = update_config(minimal_config, hyperparameters, values=values)

    assert updated_config[".lr"]["learning_rate"] == values["learning_rate"]
    assert updated_config["train"]["layers"] == values["train.layers"]
    assert updated_config["train"]["batch_size"] == values["batch_size"]


def test_update_config_with_trial(minimal_config, hyperparameters, trial):
    _, updated_config = update_config(minimal_config, hyperparameters, trial=trial)

    learning_rate = updated_config[".lr"]["learning_rate"]
    layers = updated_config["train"]["layers"]
    batch_size = updated_config["train"]["batch_size"]

    assert (
        hyperparameters["'.lr'.learning_rate"]["low"]
        <= learning_rate
        <= hyperparameters["'.lr'.learning_rate"]["high"]
    )
    assert (
        hyperparameters["train.layers"]["low"]
        <= layers
        <= hyperparameters["train.layers"]["high"]
    )
    assert layers % hyperparameters["train.layers"]["step"] == 0
    assert batch_size in hyperparameters["train.batch_size"]["choices"]


def test_update_config_raises_error_on_unknown_parameter_type(
    minimal_config, hyperparameters_with_invalid_type, trial
):
    with pytest.raises(
        ValueError,
        match="Unknown parameter type 'string' for hyperparameter 'train.optimizer'.",
    ):
        update_config(minimal_config, hyperparameters_with_invalid_type, trial=trial)


def test_update_config_raises_error_on_wrong_path(
    minimal_config, hyperparameters_with_invalid_path, trial
):
    with pytest.raises(KeyError, match="Path 'model' not found in config."):
        update_config(minimal_config, hyperparameters_with_invalid_path, trial=trial)



class StreamSentinel:
    pass


class FragmentEndSentinel(StreamSentinel):
    kind = "fragment"

    def __init__(self, name: str):
        self.name = name


class DatasetEndSentinel(StreamSentinel):
    # Singleton is important since the DatasetEndSentinel object may be passed to
    # other processes, i.e. pickled, depickled, while it should
    # always be the same object.
    kind = "dataset"
    instance = None

    def __new__(cls, *args, **kwargs):
        if cls.instance is None:
            cls.instance = super().__new__(cls)
        return cls.instance


DATASET_END_SENTINEL = DatasetEndSentinel()



import pytest

import edsnlp
from edsnlp.utils.collections import ld_to_dl

try:
    import torch.nn
except ImportError:
    torch = None


def test_map_batches():
    items = [1, 2, 3, 4, 5]
    stream = edsnlp.data.from_iterable(items)
    stream = stream.map(lambda x: x + 1)  # 2, 3, 4, 5, 6
    stream = stream.map_batches(lambda x: [sum(x)])
    stream = stream.set_processing(
        num_cpu_workers=2,
        sort_chunks=False,
        batch_size=2,
    )
    res = list(stream)
    assert res == [6, 8, 6]  # 2+4, 3+5, 6


@pytest.mark.parametrize("num_cpu_workers", [1, 2])
def test_flat_iterable(num_cpu_workers):
    items = [1, 2, 3, 4]
    stream = edsnlp.data.from_iterable(items)
    stream = stream.set_processing(num_cpu_workers=num_cpu_workers)
    stream = stream.map(lambda x: [x] * x)
    stream = stream.flatten()
    res = list(stream.to_iterable(converter=lambda x: x))
    assert sorted(res) == [1, 2, 2, 3, 3, 3, 4, 4, 4, 4]


@pytest.mark.parametrize("num_gpu_workers", [0, 1, 2])
@pytest.mark.skipif(torch is None, reason="torch not installed")
def test_map_gpu(num_gpu_workers):
    import torch

    def prepare_batch(batch, device):
        return {"tensor": torch.tensor(batch).to(device)}

    def forward(batch):
        return {"outputs": batch["tensor"] * 2}

    items = range(15)
    stream = edsnlp.data.from_iterable(items)
    if num_gpu_workers == 0:
        # this is just to fuse tests, and test map_gpu
        # following a map_batches without specifying a batch size
        stream = stream.map_batches(lambda x: x)
    stream = stream.map_gpu(prepare_batch, forward)
    stream = stream.set_processing(
        num_gpu_workers=num_gpu_workers,
        gpu_worker_devices=["cpu"] * num_gpu_workers,
        sort_chunks=False,
        batch_size=2,
    )

    res = ld_to_dl(stream)
    res = torch.cat(res["outputs"])
    assert set(res.tolist()) == {i * 2 for i in range(15)}


# fmt: off
@pytest.mark.parametrize(
    "sort,num_cpu_workers,batch_kwargs,expected",
    [
        (False, 1, {"batch_size": 10, "batch_by": "words"}, [3, 1, 3, 1, 3, 1]),  # noqa: E501
        (False, 1, {"batch_size": 10, "batch_by": "padded_words"}, [2, 1, 1, 2, 1, 1, 2, 1, 1]),  # noqa: E501
        (False, 1, {"batch_size": 10, "batch_by": "docs"}, [10, 2]),  # noqa: E501
        (False, 2, {"batch_size": 10, "batch_by": "words"}, [2, 1, 2, 1, 2, 1, 1, 1, 1]),  # noqa: E501
        (False, 2, {"batch_size": 10, "batch_by": "padded_words"}, [2, 1, 2, 1, 2, 1, 1, 1, 1]),  # noqa: E501
        (False, 2, {"batch_size": 10, "batch_by": "docs"}, [6, 6]),  # noqa: E501
        (True, 2, {"batch_size": 10, "batch_by": "padded_words"}, [3, 3, 2, 1, 1, 1, 1]),  # noqa: E501
        (False, 2, {"batch_size": "10 words"}, [2, 1, 2, 1, 2, 1, 1, 1, 1]),  # noqa: E501
    ],
)
# fmt: on
def test_map_with_batching(sort, num_cpu_workers, batch_kwargs, expected):
    nlp = edsnlp.blank("eds")
    nlp.add_pipe(
        "eds.matcher",
        config={
            "terms": {
                "foo": ["This", "is", "a", "sentence", ".", "Short", "snippet", "too"],
            }
        },
        name="matcher",
    )
    samples = [
        "This is a sentence.",
        "Short snippet",
        "Short snippet too",
        "This is a very very long sentence that will make more than 10 words",
    ] * 3
    stream = edsnlp.data.from_iterable(samples)
    if sort:
        stream = stream.map_batches(lambda x: sorted(x, key=len), batch_size=1000)
    stream = stream.map_pipeline(nlp)
    stream = stream.map_batches(len)
    stream = stream.set_processing(
        num_cpu_workers=num_cpu_workers,
        **batch_kwargs,
        chunk_size=1000,  # deprecated
        split_into_batches_after="matcher",
        show_progress=True,
    )
    assert list(stream) == expected


def test_repr(frozen_ml_nlp, tmp_path):
    items = ["ceci est un test", "ceci est un autre test"]
    stream = (
        edsnlp.data.from_iterable(items, converter=frozen_ml_nlp.make_doc)
        .map(lambda x: x)
        .map_pipeline(frozen_ml_nlp, batch_size=2)
        .map_batches(lambda b: sorted(b, key=len))
        .set_processing(num_cpu_workers=2)
        .write_json(tmp_path / "out_test.jsonl", lines=True, execute=False)
    )
    assert "Stream" in repr(stream)


@pytest.mark.parametrize("shuffle_reader", [True, False])
def test_shuffle_before_generator(shuffle_reader):
    def gen_fn(x):
        yield x
        yield x

    items = [1, 2, 3, 4, 5]
    stream = edsnlp.data.from_iterable(items)
    stream = stream.map(lambda x: x)
    stream = stream.shuffle(seed=42, shuffle_reader=shuffle_reader)
    stream = stream.map(gen_fn)
    assert stream.reader.shuffle == ("dataset" if shuffle_reader else False)
    assert len(stream.ops) == (2 if shuffle_reader else 5)
    res = list(stream)
    assert res == [4, 4, 2, 2, 3, 3, 5, 5, 1, 1]


def test_shuffle_after_generator():
    def gen_fn(x):
        yield x
        yield x

    items = [1, 2, 3, 4, 5]
    stream = edsnlp.data.from_iterable(items)
    stream = stream.map(lambda x: x)
    stream = stream.map(gen_fn)
    stream = stream.shuffle(seed=43)
    assert stream.reader.shuffle == "dataset"
    assert len(stream.ops) == 5
    res = list(stream)
    assert res == [1, 2, 4, 3, 1, 3, 5, 5, 4, 2]


def test_shuffle_frozen_ml_pipeline(run_in_test_dir, frozen_ml_nlp):
    stream = edsnlp.data.read_parquet("../resources/docs.parquet", converter="omop")
    stream = stream.map_pipeline(frozen_ml_nlp, batch_size=2)
    assert len(stream.ops) == 7
    stream = stream.shuffle(batch_by="fragment")
    assert len(stream.ops) == 7
    assert stream.reader.shuffle == "fragment"


def test_unknown_shuffle():
    items = [1, 2, 3, 4, 5]
    stream = edsnlp.data.from_iterable(items)
    stream = stream.map(lambda x: x)
    with pytest.raises(ValueError):
        stream.shuffle("unknown")


def test_int_shuffle():
    items = [1, 2, 3, 4, 5, 6, 7, 8, 9, 10]
    stream = edsnlp.data.from_iterable(items)
    stream = stream.map(lambda x: x)
    stream = stream.shuffle("2 docs", seed=42)
    assert list(stream) == [2, 1, 4, 3, 5, 6, 8, 7, 10, 9]


def test_parallel_preprocess_stop(run_in_test_dir, frozen_ml_nlp):
    nlp = frozen_ml_nlp
    stream = edsnlp.data.read_parquet(
        "../resources/docs.parquet",
        "omop",
        loop=True,
    )
    stream = stream.map(edsnlp.pipes.split(regex="\n+"))
    stream = stream.map(nlp.preprocess, kwargs=dict(supervision=True))
    stream = stream.batchify("128 words")
    stream = stream.map(nlp.collate)
    stream = stream.set_processing(num_cpu_workers=1, process_start_method="spawn")

    it = iter(stream)
    total = 0
    for _ in zip(it, range(10)):
        total += 1

    assert total == 10
    del it



from __future__ import annotations

import random
from typing import Any, Callable, Iterable, Optional, Union

import polars
import polars as pl
from typing_extensions import Literal

from edsnlp import registry
from edsnlp.core.stream import Stream
from edsnlp.data.base import BaseWriter, MemoryBasedReader
from edsnlp.data.converters import get_dict2doc_converter, get_doc2dict_converter
from edsnlp.utils.collections import flatten
from edsnlp.utils.stream_sentinels import DatasetEndSentinel
from edsnlp.utils.typing import AsList


class PolarsReader(MemoryBasedReader):
    DATA_FIELDS = ("data",)

    def __init__(
        self,
        data: Union[pl.DataFrame, pl.LazyFrame],
        shuffle: Literal["dataset", False] = False,
        seed: Optional[int] = None,
        loop: bool = False,
    ):
        super().__init__()
        self.shuffle = shuffle
        self.emitted_sentinels = {"dataset"}
        seed = seed if seed is not None else random.getrandbits(32)
        self.rng = random.Random(seed)
        self.loop = loop

        if hasattr(data, "collect"):
            data = data.collect()
        assert isinstance(data, pl.DataFrame)
        self.data = data

    def read_records(self) -> Iterable[Any]:
        data: polars.DataFrame = self.data
        while True:
            if self.shuffle:
                data = self.data.sample(
                    fraction=1.0,
                    seed=self.rng.getrandbits(32),
                    shuffle=True,
                )
            yield from data.iter_rows(named=True)
            yield DatasetEndSentinel()
            if not self.loop:
                break

    def __repr__(self):
        return (
            f"{self.__class__.__name__}(data={object.__repr__(self.data)}, "
            f"shuffle={self.shuffle}, "
            f"loop={self.loop})"
        )


@registry.readers.register("polars")
def from_polars(
    data: Union[pl.DataFrame, pl.LazyFrame],
    shuffle: Literal["dataset", False] = False,
    seed: Optional[int] = None,
    loop: bool = False,
    converter: Optional[AsList[Union[str, Callable]]] = None,
    **kwargs,
) -> Stream:
    """
    The PolarsReader (or `edsnlp.data.from_polars`) handles reading from a table and
    yields documents. At the moment, only entities and attributes are loaded. Relations
    and events are not supported.

    Example
    -------
    ```{ .python .no-check }

    import edsnlp

    nlp = edsnlp.blank("eds")
    nlp.add_pipe(...)
    doc_iterator = edsnlp.data.from_polars(df, nlp=nlp, converter="omop")
    annotated_docs = nlp.pipe(doc_iterator)
    ```

    !!! note "Generator vs list"

        `edsnlp.data.from_polars` returns a
        [Stream][edsnlp.core.stream.Stream].
        To iterate over the documents multiple times efficiently or to access them by
        index, you must convert it to a list

        ```{ .python .no-check }
        docs = list(edsnlp.data.from_polars(df, converter="omop"))
        ```

    Parameters
    ----------
    data: Union[pl.DataFrame, pl.LazyFrame]
        Polars object
    shuffle: Literal["dataset", False]
        Whether to shuffle the data. If "dataset", the whole dataset will be shuffled
        at the beginning (of every epoch if looping).
    seed: int
        The seed to use for shuffling.
    loop: bool
        Whether to loop over the data indefinitely.
    converter: Optional[AsList[Union[str, Callable]]]
        Converters to use to convert the rows of the DataFrame (represented as dicts)
        to Doc objects. These are documented on the [Converters](/data/converters) page.
    kwargs:
        Additional keyword arguments to pass to the converter. These are documented on
        the [Converters](/data/converters) page.

    Returns
    -------
    Stream
    """

    data = Stream(
        reader=PolarsReader(
            data,
            shuffle=shuffle,
            seed=seed,
            loop=loop,
        )
    )
    if converter:
        for conv in converter:
            conv, kwargs = get_dict2doc_converter(conv, kwargs)
            data = data.map(conv, kwargs=kwargs)
    return data


class PolarsWriter(BaseWriter):
    def __init__(self, dtypes: Optional[dict] = None):
        self.dtypes = dtypes

    def consolidate(self, items: Iterable[Any]):
        return pl.from_dicts(flatten(items), schema=self.dtypes)


@registry.writers.register("polars")
def to_polars(
    data: Union[Any, Stream],
    converter: Optional[Union[str, Callable]] = None,
    dtypes: Optional[dict] = None,
    execute: bool = True,
    **kwargs,
) -> pl.DataFrame:
    """
    `edsnlp.data.to_polars` writes a list of documents as a polars dataframe.

    Example
    -------
    ```{ .python .no-check }

    import edsnlp

    nlp = edsnlp.blank("eds")
    nlp.add_pipe(...)

    doc = nlp("My document with entities")

    edsnlp.data.to_polars([doc], converter="omop")
    ```

    Parameters
    ----------
    data: Union[Any, Stream],
        The data to write (either a list of documents or a Stream).
    dtypes: Optional[dict]
        Dictionary of column names to dtypes. This is passed to the schema parameter of
        `pl.from_dicts`.
    converter: Optional[Union[str, Callable]]
        Converter to use to convert the documents to dictionary objects before storing
        them in the dataframe. These are documented on the
        [Converters](/data/converters) page.
    execute: bool
        Whether to execute the writing operation immediately or to return a stream
    kwargs:
        Additional keyword arguments to pass to the converter. These are documented on
        the [Converters](/data/converters) page.
    """
    data = Stream.ensure_stream(data)
    if converter:
        converter, kwargs = get_doc2dict_converter(converter, kwargs)
        data = data.map(converter, kwargs=kwargs)

    return data.write(PolarsWriter(dtypes), execute=execute)



import asyncio
import threading
from typing import Any, Coroutine, Dict, Iterable, Optional, Tuple


class AsyncRequestWorker:
    _instance = None

    def __init__(self):
        self.loop = asyncio.new_event_loop()
        self.thread = threading.Thread(target=self._run, daemon=True)
        self.thread.start()
        self._lock = threading.Lock()
        self._cv = threading.Condition(self._lock)
        self._next_id = 0
        self._results: Dict[int, Tuple[Any, Optional[BaseException]]] = {}

    def _run(self):
        asyncio.set_event_loop(self.loop)
        self.loop.run_forever()

    @classmethod
    def instance(cls) -> "AsyncRequestWorker":
        if cls._instance is None:
            cls._instance = AsyncRequestWorker()
        return cls._instance

    def submit(self, coro: Coroutine[Any, Any, Any]) -> int:
        with self._lock:
            task_id = self._next_id
            self._next_id += 1

        async def _wrap():
            try:
                res = await coro
                exc = None
            except BaseException as e:  # noqa: BLE001
                res = None
                exc = e
            with self._cv:
                self._results[task_id] = (res, exc)
                self._cv.notify_all()

        asyncio.run_coroutine_threadsafe(_wrap(), self.loop)
        return task_id

    def pop_result(self, task_id: int) -> Optional[Tuple[Any, Optional[BaseException]]]:
        with self._cv:
            return self._results.pop(task_id, None)

    def wait_for_any(self, task_ids: Iterable[int]) -> int:
        task_ids = set(task_ids)
        with self._cv:
            while True:
                for tid in task_ids:
                    if tid in self._results:
                        return tid
                self._cv.wait()



from typing import Iterable, List, Optional, Tuple

from spacy.tokens import Doc, Span

import edsnlp
from edsnlp import Pipeline
from edsnlp.utils.span_getters import SpanGetterArg, get_spans_with_group


@edsnlp.registry.factory.register("eds.explode", spacy_compatible=False)
class Explode:
    def __init__(
        self,
        nlp: Optional[Pipeline] = None,
        name: str = "explode",
        *,
        span_getter: SpanGetterArg = {"ents": True},
        filter_expr: Optional[str] = None,
    ):
        """
        Explode a Doc into multiple distinct Doc objects, one per span retrieved through
        the `span_getter` : each span becomes alone in its own Doc. Note that entities
        that are not selected by the `span_getter` will be lost in the new docs.

        !!! warning "Not for pipelines"

            This component is not meant to be used in a pipeline, but rather
            as a preprocessing step when dealing with a stream of documents
            as in the example below.

        !!! note "Difference with `eds.split`"

            While `eds.split` breaks a document into smaller chunks based on length or
            regex rules, `eds.explode` creates a separate document for each selected
            span. This means `eds.split` is typically used for segmenting text for
            context size or processing constraints, whereas `eds.explode` is designed
            for span-level tasks that require span-level mixing, like training span
            classifiers, ensuring that each span is isolated in its own document while
            preserving the original context.

        Examples
        --------
        ```python
        import edsnlp.pipes as eds
        from edsnlp.data.converters import MarkupToDocConverter

        converter = MarkupToDocConverter(
            preset="xml",
            # Put xml annotated spans in distinct doc.spans[label] groups
            span_setter={"*": True},
        )
        doc = converter(
            "Le <person>patient</person> a mal au <body_part>bras</body_part>, à la "
            "<body_part>jambe</body_part> et au <body_part>torse</body_part>"
        )

        exploder = eds.explode(span_getter=["body_part"])
        print(doc.text, "->", doc.spans)
        # Out: Le patient a mal au bras, à la jambe et au torse -> {'person': [patient], 'body_part': [bras, jambe, torse]}

        for new_doc in exploder(doc):
            print(new_doc.text, "->", new_doc.spans)
        # Out: Le patient a mal au bras, à la jambe et au torse -> {'person': [], 'body_part': [bras]}
        # Out: Le patient a mal au bras, à la jambe et au torse -> {'person': [], 'body_part': [jambe]}
        # Out: Le patient a mal au bras, à la jambe et au torse -> {'person': [], 'body_part': [torse]}
        ```

        Parameters
        ----------
        nlp: Optional[Pipeline]
            The pipeline object
        name: str
            Name of the pipe
        span_getter: SpanGetterArg
            The span getter to use to retrieve spans from the Doc.
            Default is `{"ents": True}` which retrieves all entities in `doc.ents`.
        filter_expr: Optional[str]
            An optional filter expression to filter the produced documents. The callable
            expects a single argument, the new Doc, and should return a boolean.
        """  # noqa: E501
        self.span_getter = span_getter
        self.filter_fn = eval(f"lambda doc:{filter_expr}") if filter_expr else None

    def __call__(self, doc: Doc) -> Iterable[Doc]:
        for new_doc in self.explode_doc(doc):
            if not self.filter_fn or self.filter_fn(new_doc):
                yield new_doc

    def explode_doc(self, doc: Doc) -> Iterable[Doc]:
        """
        Yield a sequence of docs, one per span returned by the getter.
        """
        base = doc.copy()
        span_pairs: List[Tuple[Span, str]] = list(
            get_spans_with_group(base, self.span_getter)
        )
        # Nothing to explode: return the original doc unchanged
        if not span_pairs:
            yield doc
            return

        for name in base.spans:
            base.spans[name] = []
        base.ents = []

        for span, group in span_pairs:
            if group == "ents":
                base.ents = [span]
                new_doc = base.copy()
                base.ents = []
            else:
                base.spans[group] = [span]
                new_doc = base.copy()
                base.spans[group] = []

            yield new_doc



import re
from typing import TYPE_CHECKING, Any, List, Optional, Union

import regex
from pydantic import BaseModel

from edsnlp.matchers.utils import ListOrStr
from edsnlp.utils.span_getters import (
    ContextWindow,
    SentenceContextWindow,
    SpanGetterArg,
)
from edsnlp.utils.typing import AsList

Flags = Union[re.RegexFlag, int]

try:
    from pydantic import field_validator, model_validator

    def validator(x, allow_reuse=True, pre=False):
        return field_validator(x, mode="before" if pre else "after")

    def root_validator(allow_reuse=True, pre=False):
        return model_validator(mode="before" if pre else "after")


except ImportError:
    from pydantic import root_validator, validator


def validate_window(cls, values):
    if isinstance(values.get("regex"), str):
        values["regex"] = [values["regex"]]
    window = values.get("window")
    if window is None or isinstance(window, (int, tuple, list)):
        values["limit_to_sentence"] = True
    window = values.get("window")
    if window is not None:
        values["window"] = ContextWindow.validate(window)
    if values.get("limit_to_sentence"):
        values["window"] = (
            SentenceContextWindow(0, 0) & values.get("window")
            if window is not None
            else SentenceContextWindow(0, 0)
        )
    return values


class SingleExcludeModel(BaseModel):
    """
    A dictionary to define exclusion rules. Exclusion rules are given as Regexes, and
    if a match is found in the surrounding context of an extraction, the extraction is
    removed. Note that only take a match into account if it is not inside the anchor
    span.

    Parameters
    ----------
    regex : ListOrStr
        A single Regex or a list of Regexes
    regex_attr : Optional[str]
        An attributes to overwrite the given `attr` when matching with Regexes.
    regex_flags : re.RegexFlag
        Regex flags
    span_getter : Optional[SpanGetterArg]
        A span getter to pick the assigned spans from already extracted entities.
    window : Optional[ContextWindow]
        Context window to search for patterns around the anchor. Defaults to "sent" (
        i.e. the sentence of the anchor span).
    """

    span_getter: Optional[SpanGetterArg] = None
    regex: ListOrStr = []
    regex_attr: Optional[str] = None
    regex_flags: Union[re.RegexFlag, int] = None

    limit_to_sentence: Optional[bool] = None
    window: Optional[ContextWindow] = None
    regex_matcher: Optional[Any] = None

    validate_window = root_validator(pre=True, allow_reuse=True)(validate_window)


class SingleIncludeModel(BaseModel):
    """
    A dictionary to define inclusion rules. Inclusion rules are given as Regexes, and
    if a match isn't found in the surrounding context of an extraction, the extraction
    is removed. Note that only take a match into account if it is not inside the anchor
    span.

    Parameters
    ----------
    regex : ListOrStr
        A single Regex or a list of Regexes
    regex_attr : Optional[str]
        An attributes to overwrite the given `attr` when matching with Regexes.
    regex_flags : re.RegexFlag
        Regex flags
    span_getter : Optional[SpanGetterArg]
        A span getter to pick the assigned spans from already extracted entities.
    window : Optional[ContextWindow]
        Context window to search for patterns around the anchor. Defaults to "sent" (
        i.e. the sentence of the anchor span).
    """

    span_getter: Optional[SpanGetterArg] = None
    regex: ListOrStr = []
    regex_attr: Optional[str] = None
    regex_flags: Union[re.RegexFlag, int] = None

    limit_to_sentence: Optional[bool] = None
    window: Optional[ContextWindow] = None

    regex_matcher: Optional[Any] = None

    validate_window = root_validator(pre=True, allow_reuse=True)(validate_window)


class ExcludeModel(AsList[SingleExcludeModel]):
    """
    A list of `SingleExcludeModel` objects. If a single config is passed,
    it will be automatically converted to a list of a single element.
    """


class IncludeModel(AsList[SingleIncludeModel]):
    """
    A list of `SingleIncludeModel` objects. If a single config is passed,
    it will be automatically converted to a list of a single element.
    """


class SingleAssignModel(BaseModel):
    """
    A dictionary to refine the extraction. Similarly to the `exclude` key, you can
    provide a dictionary to use on the context **before** and **after** the extraction.

    Parameters
    ----------
    span_getter : Optional[SpanGetterArg]
        A span getter to pick the assigned spans from already extracted entities
        in the doc.
    regex : ListOrStr
        A single Regex or a list of Regexes
    regex_attr : Optional[str]
        An attributes to overwrite the given `attr` when matching with Regexes.
    regex_flags : re.RegexFlag
        Regex flags
    window : Optional[ContextWindow]
        Context window to search for patterns around the anchor. Defaults to "sent" (
        i.e. the sentence of the anchor span).
    replace_entity : Optional[bool]
        If set to `True`, the match from the corresponding assign key will be used as
        entity, instead of the main match.
        See [this paragraph][replace_entity]
    reduce_mode : Optional[Flags]
        Set how multiple assign matches are handled. See the documentation of the
        [`reduce_mode` parameter][reduce_mode]
    required : Optional[str]
        If set to `True`, the assign key must match for the extraction to be kept. If
        it does not match, the extraction is discarded.
    name : str
        A name (string)
    """

    name: str

    span_getter: Optional[SpanGetterArg] = None
    regex: ListOrStr = []
    regex_attr: Optional[str] = None
    regex_flags: Union[re.RegexFlag, int] = None

    limit_to_sentence: Optional[bool] = None
    window: Optional[ContextWindow] = None
    replace_entity: bool = False
    reduce_mode: Optional[str] = None
    required: Optional[bool] = False

    regex_matcher: Optional[Any] = None

    @validator("regex", allow_reuse=True)
    def check_single_regex_group(cls, pat):
        for single_pat in pat:
            compiled_pat = regex.compile(
                single_pat
            )  # Using regex to allow multiple fgroups with same name
            n_groups = compiled_pat.groups
            assert n_groups == 1, (
                f"The pattern {single_pat} should have exactly one capturing group, "
                f"not {n_groups}"
            )

        return pat

    validate_window = root_validator(pre=True, allow_reuse=True)(validate_window)


class AssignModel(AsList[SingleAssignModel]):
    """
    A list of `SingleAssignModel` objects that should have at most
    one element with `replace_entity=True`. If a single config is passed,
    it will be automatically converted to a list of a single element.
    """

    @classmethod
    def name_uniqueness(cls, v, config=None):
        names = [item.name for item in v]
        assert len(names) == len(set(names)), "Each `name` field should be unique"
        return v

    @classmethod
    def replace_uniqueness(cls, v, config=None):
        replace = [item for item in v if item.replace_entity]
        assert len(replace) <= 1, (
            "Only 1 assign element can be set with `replace_entity=True`"
        )
        return v

    @classmethod
    def __get_validators__(cls):
        yield cls.validate
        yield cls.name_uniqueness
        yield cls.replace_uniqueness


if TYPE_CHECKING:
    ExcludeModel = List[SingleExcludeModel]  # noqa: F811
    IncludeModel = List[SingleIncludeModel]  # noqa: F811
    AssignModel = List[SingleAssignModel]  # noqa: F811


class SingleConfig(BaseModel, extra="forbid"):
    """
    A single configuration for the contextual matcher.

    Parameters
    ----------
    span_getter : Optional[SpanGetterArg]
        A span getter to pick the assigned spans from already extracted entities
        in the doc.
    regex : ListOrStr
        A single Regex or a list of Regexes
    regex_attr : Optional[str]
        An attributes to overwrite the given `attr` when matching with Regexes.
    regex_flags: re.RegexFlag
        Regex flags
    terms : Union[re.RegexFlag, int]
        A single term or a list of terms (for exact matches)
    exclude : AsList[SingleExcludeModel]
        ??? subdoc "One or more exclusion patterns"

            ::: edsnlp.pipes.core.contextual_matcher.models.SingleExcludeModel
                options:
                    only_parameters: "no-header"
    include : AsList[SingleIncludeModel]
        ??? subdoc "One or more inclusion patterns"

            ::: edsnlp.pipes.core.contextual_matcher.models.SingleIncludeModel
                options:
                    only_parameters: "no-header"
    assign : AsList[SingleAssignModel]
        ??? subdoc "One or more assignment patterns"

            ::: edsnlp.pipes.core.contextual_matcher.models.SingleAssignModel
                options:
                    only_parameters: "no-header"
    source : str
        A label describing the pattern

    """

    source: Optional[str] = None

    span_getter: Optional[SpanGetterArg] = None
    terms: ListOrStr = []
    regex: ListOrStr = []
    regex_attr: Optional[str] = None
    regex_flags: Union[re.RegexFlag, int] = None

    exclude: ExcludeModel = []
    include: IncludeModel = []
    assign: AssignModel = []

    regex_matcher: Optional[Any] = None
    phrase_matcher: Optional[Any] = None


class FullConfig(AsList[SingleConfig]):
    """
    A list of `SingleConfig` objects that should have distinct `source` fields.
    If a single config is passed, it will be automatically converted to a list of
    a single element.
    """

    @classmethod
    def source_uniqueness(cls, v, config=None):
        sources = [item.source for item in v]
        assert len(sources) == len(set(sources)), "Each `source` field should be unique"
        return v

    @classmethod
    def __get_validators__(cls):
        yield cls.validate
        yield cls.source_uniqueness


if TYPE_CHECKING:
    FullConfig = List[SingleConfig]  # noqa: F811



import random
import warnings
from pathlib import Path
from typing import Any, Callable, Dict, Iterable, List, Optional, Tuple, Union

import pyarrow.dataset
import pyarrow.fs
import pyarrow.parquet
from confit import VisibleDeprecationWarning
from pyarrow.dataset import ParquetFileFragment
from typing_extensions import Literal

from edsnlp import registry
from edsnlp.core.stream import Stream
from edsnlp.data.base import BatchWriter, FileBasedReader
from edsnlp.data.converters import get_dict2doc_converter, get_doc2dict_converter
from edsnlp.utils.batching import BatchBy, batchify_fns
from edsnlp.utils.collections import batchify, dl_to_ld, flatten, ld_to_dl, shuffle
from edsnlp.utils.file_system import FileSystem, normalize_fs_path
from edsnlp.utils.stream_sentinels import DatasetEndSentinel, FragmentEndSentinel
from edsnlp.utils.typing import AsList


class ParquetReader(FileBasedReader):
    DATA_FIELDS = ()

    def __init__(
        self,
        path: Union[str, Path],
        *,
        filesystem: Optional[FileSystem] = None,
        shuffle: Literal["dataset", "fragment", False] = False,
        seed: Optional[int] = None,
        loop: bool = False,
        work_unit: Literal["record", "fragment"] = "record",
    ):
        super().__init__()
        self.shuffle = shuffle
        self.emitted_sentinels = {"dataset"} | (
            set() if shuffle == "dataset" else {"fragment"}
        )
        seed = seed if seed is not None else random.getrandbits(32)
        self.rng = random.Random(seed)
        self.loop = loop
        self.work_unit = work_unit
        assert not (work_unit == "fragment" and shuffle == "dataset"), (
            "Cannot shuffle at the dataset level and dispatch tasks at the "
            "fragment level. Set shuffle='fragment' or work_unit='record'."
        )
        # Either the filesystem has not been passed
        # or the path is a URL (e.g. s3://) => we need to infer the filesystem
        self.fs, self.path = normalize_fs_path(filesystem, path)
        self.fragments: List[ParquetFileFragment] = list(
            pyarrow.dataset.dataset(
                self.path,
                format="parquet",
                filesystem=self.fs,
            ).get_fragments()
        )

    def read_fragment(self, fragment: ParquetFileFragment) -> Iterable[Dict]:
        return (
            doc
            for batch in fragment.scanner().to_reader()
            for doc in dl_to_ld(batch.to_pydict())
        )

    def extract_task(self, item):
        if self.work_unit == "fragment":
            records = self.read_fragment(item)
            if self.shuffle == "fragment":
                records = shuffle(records, self.rng)
            yield from records
        else:
            yield item

    def read_records(self) -> Iterable[Any]:
        while True:
            files = self.fragments
            if self.shuffle:
                files = shuffle(files, self.rng)
            if self.shuffle == "fragment":
                for file in files:
                    if self.work_unit == "fragment":
                        yield file
                    else:
                        yield from shuffle(self.read_fragment(file), self.rng)
                    yield FragmentEndSentinel(file.path)
            elif self.shuffle == "dataset":
                assert self.work_unit == "record"
                records = (line for file in files for line in self.read_fragment(file))
                yield from shuffle(records, self.rng)
            else:
                for file in files:
                    if self.work_unit == "fragment":
                        yield file
                    else:
                        yield from self.read_fragment(file)
                    yield FragmentEndSentinel(file.path)
            yield DatasetEndSentinel()
            if not self.loop:
                break

    def __repr__(self):
        return (
            f"{self.__class__.__name__}("
            f"path={self.path!r}, "
            f"shuffle={self.shuffle}, "
            f"loop={self.loop})"
        )


class ParquetWriter(BatchWriter):
    def __init__(
        self,
        *,
        path: Union[str, Path],
        batch_size: Optional[Union[int, str]] = None,
        batch_by: BatchBy = None,
        write_in_worker: bool = False,
        overwrite: bool,
        filesystem: Optional[FileSystem] = None,
        pyarrow_write_kwargs: Optional[Dict[str, Any]] = None,
    ):
        super().__init__()
        self.fs, self.path = normalize_fs_path(filesystem, path)

        # Check that filesystem has the same protocol as indicated by path
        self.fs.makedirs(self.path, exist_ok=True)
        self.pyarrow_write_kwargs = pyarrow_write_kwargs or {}

        dataset: pyarrow.dataset.FileSystemDataset = (  # type: ignore
            pyarrow.dataset.dataset(
                self.path,
                format="parquet",
                filesystem=self.fs,
            )
        )
        if len(list(dataset.files)):
            if not overwrite:
                raise FileExistsError(
                    f"Directory {self.path} already exists and is not empty. "
                    "Use overwrite=True to overwrite."
                )
            for file in dataset.files:
                self.fs.rm_file(file)
        self.fs = filesystem
        batch_size, batch_by = Stream.validate_batching(batch_size, batch_by)
        if batch_by in ("docs", "doc", None, batchify) and batch_size is None:
            warnings.warn(
                "You should specify a batch size when using record-wise batch writing. "
                "Setting batch size to 1024."
            )
            batch_size = 1024
        batch_by = batch_by or "docs"
        self.batch_fn = batchify_fns.get(batch_by, batch_by)

        self.batch_size = batch_size
        self.write_in_worker = write_in_worker
        self.batch = []
        self.closed = False

    def handle_batch(self, batch: List[Dict]) -> Tuple[ParquetFileFragment, int]:
        fragment = pyarrow.Table.from_pydict(ld_to_dl(flatten(batch)))
        pyarrow.parquet.write_to_dataset(
            table=fragment,
            root_path=self.path,
            filesystem=self.fs,
            **self.pyarrow_write_kwargs,
        )
        return (None, len(batch))

    def consolidate(
        self, items: Iterable[ParquetFileFragment]
    ) -> pyarrow.dataset.Dataset:
        for _ in items:
            pass
        return pyarrow.dataset.dataset(self.path, format="parquet", filesystem=self.fs)


@registry.readers.register("parquet")
def read_parquet(
    path: Union[str, Path],
    converter: Optional[AsList[Union[str, Callable]]] = None,
    *,
    filesystem: Optional[FileSystem] = None,
    shuffle: Literal["dataset", "fragment", False] = False,
    seed: Optional[int] = None,
    loop: bool = False,
    work_unit: Literal["record", "fragment"] = "record",
    **kwargs,
) -> Stream:
    """
    The ParquetReader (or `edsnlp.data.read_parquet`) reads a directory of parquet files
    (or a single file) and yields documents.

    Example
    -------
    ```{ .python .no-check }

    import edsnlp

    nlp = edsnlp.blank("eds")
    nlp.add_pipe(...)
    doc_iterator = edsnlp.data.read_parquet("path/to/parquet", converter="omop")
    annotated_docs = nlp.pipe(doc_iterator)
    ```

    !!! note "Generator vs list"

        `edsnlp.data.read_parquet` returns a
        [Stream][edsnlp.core.stream.Stream].
        To iterate over the documents multiple times efficiently or to access them by
        index, you must convert it to a list

        ```{ .python .no-check }
        docs = list(edsnlp.data.read_parquet("path/to/parquet", converter="omop"))
        ```

    Parameters
    ----------
    path: Union[str, Path]
        Path to the directory containing the parquet files (will recursively look for
        files in subdirectories). Supports any filesystem supported by pyarrow.
    filesystem: Optional[AbstractFileSystem] = None,
        The filesystem to use to write the files. If None, the filesystem will be
        inferred from the path (e.g. `s3://` will use S3).
    shuffle: Literal["dataset", "fragment", False]
        Whether to shuffle the data. If "dataset", the whole dataset will be shuffled
        before starting iterating on it (at the start of every epoch if looping). If
        "fragment", shuffling will occur between and inside the parquet files, but not
        across them.

        !!! warning "Dataset shuffling"

            Shuffling the dataset can be expensive, especially for large datasets,
            since it requires reading the entire dataset into memory. If you have a
            large dataset, consider shuffling at the "fragment" level.
    seed: Optional[int]
        The seed to use for shuffling.
    loop: bool
        Whether to loop over the data indefinitely.
    work_unit: Literal["record", "fragment"]
        Only affects the multiprocessing mode. If "record", every worker will start to
        read the same parquet file and yield each every num_workers-th record, starting
        at an offset each. For instance, if num_workers=2, the first worker will read
        the 1st, 3rd, 5th, ... records, while the second worker will read the 2nd, 4th,
        6th, ... records of the first parquet file.

        If "fragment", each worker will read a different parquet file. For instance, the
        first worker will every record of the 1st parquet file, the second worker will
        read every record of the 2nd parquet file, and so on. This way, no record is
        "wasted" and every record loaded in memory is yielded.

    converter: Optional[AsList[Union[str, Callable]]]
        Converters to use to convert the parquet rows of the data source to Doc objects
        These are documented on the [Converters](/data/converters) page.
    kwargs:
        Additional keyword arguments to pass to the converter. These are documented on
        the [Converters](/data/converters) page.

    Returns
    -------
    Stream
    """
    if "read_in_worker" in kwargs:
        warnings.warn(
            "The `read_in_worker` parameter of edsnlp.data.read_parquet is deprecated "
            "and set to True by default.",
            FutureWarning,
        )
        kwargs.pop("read_in_worker")

    data = Stream(
        reader=ParquetReader(
            path,
            filesystem=filesystem,
            shuffle=shuffle,
            seed=seed,
            loop=loop,
            work_unit=work_unit,
        )
    )
    if converter:
        for conv in converter:
            conv, kw = get_dict2doc_converter(conv, kwargs)
            data = data.map(conv, kwargs=kw)
    return data


@registry.writers.register("parquet")
def write_parquet(
    data: Union[Any, Stream],
    path: Union[str, Path],
    *,
    batch_size: Optional[Union[int, str]] = None,
    batch_by: BatchBy = None,
    write_in_worker: bool = True,
    overwrite: bool = False,
    filesystem: Optional[FileSystem] = None,
    execute: bool = True,
    converter: Optional[Union[str, Callable]] = None,
    pyarrow_write_kwargs: Optional[Dict[str, Any]] = None,
    **kwargs,
) -> pyarrow.dataset.Dataset:
    """
    `edsnlp.data.write_parquet` writes a list of documents as a parquet dataset.

    Example
    -------
    ```{ .python .no-check }

    import edsnlp

    nlp = edsnlp.blank("eds")
    nlp.add_pipe(...)

    doc = nlp("My document with entities")

    edsnlp.data.write_parquet([doc], "path/to/parquet")
    ```

    !!! warning "Overwriting files"

        By default, `write_parquet` will raise an error if the directory already exists
        and contains parquet files. This is to avoid overwriting existing annotations.
        To allow overwriting existing files, use `overwrite=True`.

    Parameters
    ----------
    data: Union[Any, Stream],
        The data to write (either a list of documents or a Stream).
    path: Union[str, Path]
        Path to the directory containing the parquet files (will recursively look for
        files in subdirectories). Supports any filesystem supported by pyarrow.
    batch_size: Optional[int]
        The maximum number of documents to write in each parquet file.
    batch_by: Union[Callable, Literal["docs", "fragment"]]
        The method to batch the documents. If "docs", the batch size is the number of
        documents. If "fragment", each batch corresponds to a parquet file fragment from
        the input data.
    write_in_worker: bool
        In multiprocessing or spark mode, whether to batch and write the documents in
        the workers or in the main process.

        For instance, a worker may read the 1st, 3rd, 5th, ... documents, while another
        reads the 2nd, 4th, 6th, ... documents.

        If `write_in_worker` is False, `deterministic` is True (default) and no
        operation adds or remove document from the stream (e.g., no `map_batches`), the
        original order of the documents will be recovered in the main process, and
        batching there can produce fragments that respect the original order.
    overwrite: bool
        Whether to overwrite existing directories.
    filesystem: Optional[AbstractFileSystem] = None,
        The filesystem to use to write the files. If None, the filesystem will be
        inferred from the path (e.g. `s3://` will use S3).
    pyarrow_write_kwargs: Optional[Dict[str, Any]]
        Additional keyword arguments to pass to the `pyarrow.parquet.write_to_dataset`
    execute: bool
        Whether to execute the writing operation immediately or to return a stream
    converter: Optional[Union[str, Callable]]
        Converter to use to convert the documents to dictionary objects before writing
        them as Parquet rows. These are documented on the [Converters](/data/converters)
        page.
    kwargs:
        Additional keyword arguments to pass to the converter. These are documented on
        the [Converters](/data/converters) page.
    """

    data = Stream.ensure_stream(data)
    if "num_rows_per_file" in kwargs:
        assert batch_size is None, (
            "Cannot specify both 'batch_size' and deprecated 'num_rows_per_file'."
        )
        batch_size = kwargs.pop("num_rows_per_file")
        assert batch_by in (
            None,
            "docs",
        ), "Cannot use 'num_rows_per_file' with 'batch_by'."
    if "accumulate" in kwargs:
        warnings.warn(
            "The 'accumulate' parameter is deprecated.", VisibleDeprecationWarning
        )
    if converter:
        converter, kw = get_doc2dict_converter(converter, kwargs)
        data = data.map(converter, kwargs=kw)

    return data.write(
        ParquetWriter(
            path=path,
            batch_size=batch_size,
            batch_by=batch_by,
            write_in_worker=write_in_worker,
            pyarrow_write_kwargs=pyarrow_write_kwargs,
            overwrite=overwrite,
            filesystem=filesystem,
        ),
        execute=execute,
    )



# ruff:noqa:E402
import pytest

try:
    import torch.nn
except ImportError:
    torch = None

if torch is None:
    pytest.skip("torch not installed", allow_module_level=True)
pytest.importorskip("rich")

from edsnlp.training.optimizer import LinearSchedule, ScheduledOptimizer


class Net(torch.nn.Module):
    def __init__(self):
        super(Net, self).__init__()
        self.fc1 = torch.nn.Linear(10, 1)
        self.fc2 = torch.nn.Linear(1, 1)

    def forward(self, x):
        return self.fc(x)


@pytest.fixture(scope="module")
def net():
    net = Net()
    return net


@pytest.mark.parametrize(
    "groups",
    [
        # Old schedule API
        {
            "fc1[.].*": {
                "lr": 0.1,
                "weight_decay": 0.01,
                "schedules": [
                    {
                        "@schedules": "linear",
                        "start_value": 0.0,
                        "warmup_rate": 0.2,
                    },
                ],
            },
            "fc2[.]bias": False,
            "": {
                "lr": 0.0001,
                "weight_decay": 0.0,
            },
        },
        # New schedule API
        {
            "fc1[.].*": {
                "lr": {
                    "@schedules": "linear",
                    "start_value": 0.0,
                    "max_value": 0.1,
                    "warmup_rate": 0.2,
                },
                "weight_decay": 0.01,
            },
            "fc2[.]bias": False,
            "": {
                "lr": 0.0001,
                "weight_decay": 0.0,
            },
        },
        # New groups format
        [
            {
                "selector": "fc1[.].*",
                "lr": {
                    "@schedules": "linear",
                    "start_value": 0.0,
                    "max_value": 0.1,
                    "warmup_rate": 0.2,
                },
                "weight_decay": 0.01,
            },
            {
                "selector": "fc2[.]bias",
                "exclude": True,
            },
            {
                "selector": "",
                "lr": 0.0001,
                "weight_decay": 0.0,
            },
        ],
    ],
)
def test_old_parameter_selection(net, groups):
    optim = ScheduledOptimizer(
        optim="adamw",
        module=net,
        groups=groups,
        total_steps=10,
    )
    assert len(optim.state) == 0
    optim.initialize()
    assert all([p in optim.state for p in net.fc1.parameters()])
    optim.state = optim.state

    fc1_group = optim.param_groups[1]
    assert fc1_group["lr"] == pytest.approx(0.0)
    assert fc1_group["weight_decay"] == pytest.approx(0.01)
    assert set(fc1_group["params"]) == {net.fc1.weight, net.fc1.bias}

    fc2_group = optim.param_groups[0]
    assert fc2_group["lr"] == pytest.approx(0.0001)
    assert set(fc2_group["params"]) == {net.fc2.weight}

    lr_values = [fc1_group["lr"]]

    for i in range(10):
        optim.step()
        lr_values.append(fc1_group["lr"])

    assert lr_values == pytest.approx(
        [
            0.0,
            0.05,
            0.1,
            0.0875,
            0.075,
            0.0625,
            0.05,
            0.0375,
            0.025,
            0.0125,
            0.0,
        ]
    )


def test_serialization(net):
    optim = ScheduledOptimizer(
        optim="adamw",
        module=net,
        groups={
            "fc1[.].*": {
                "lr": 0.1,
                "weight_decay": 0.01,
                "schedules": LinearSchedule(start_value=0.0, warmup_rate=0.2),
            },
            "fc2[.]bias": False,
            "": {
                "lr": 0.0001,
                "weight_decay": 0.0,
            },
        },
        total_steps=10,
    )
    optim.initialize()
    optim.param_groups = optim.param_groups

    state_dict = None
    for i in range(10):
        if i == 5:
            state_dict = optim.state_dict()
        optim.step()

    assert optim.param_groups[-1]["lr"] == pytest.approx(0.0)
    optim.load_state_dict(state_dict)
    assert optim.param_groups[-1]["lr"] == pytest.approx(0.0625)

    optim.reset()


def test_repr(net):
    optim = ScheduledOptimizer(
        optim="adamw",
        module=net,
        groups={
            "fc1[.].*": {
                "lr": 0.1,
                "weight_decay": 0.01,
                "schedules": [
                    LinearSchedule(start_value=0.0, warmup_rate=0.2),
                    LinearSchedule(path="weight_decay"),
                ],
            },
            "fc2[.]bias": False,
            ".*": {
                "lr": 0.0001,
                "weight_decay": 0.0,
            },
        },
        total_steps=10,
    )
    optim.initialize()

    assert "ScheduledOptimizer[AdamW]" in repr(optim)


def test_warn_empty_selector(net):
    with pytest.warns(
        UserWarning,
        match="Selectors ['fc3[.].*'] did not match any parameters.",
    ):
        ScheduledOptimizer(
            optim="adamw",
            module=net,
            groups=[
                {
                    "selector": "fc3[.].*",
                    "lr": 0.1,
                    "weight_decay": 0.01,
                    "schedules": LinearSchedule(start_value=0.0, warmup_rate=0.2),
                },
                {"selector": "fc2[.]bias", "exclude": True},
                {"selector": "", "lr": 0.0001, "weight_decay": 0.0},
            ],
            total_steps=10,
        )



patterns = {
    "suicide_attempt_unspecific": [
        r"\b(?<!\.)(?<!Voie\s\d\s\:\s)(?<!Voie\sd.abord\s\:\s)(?<!surface\s)(?<!d[ée]sorientation\s)(?<!abord\s)(?<!ECG\s:\s)(?<!volume\s)(?<!\d\s[mc]m\sde\sla\s)(?<!\d[mc]m\sde\sla\s)(?<!au\scontact\sde\sla\s)T\.?S\.?(?![\.A-Za-z])(?!\sapyr[eé]tique)(?!.+TRANSSEPTAL)(?!.+T[34])(?!.+en\sr.gression)\b",
        r"(?<!\.)T\.S\.(?![A-Za-z])",
        r"\b(?<!.)TS\.\B",
        r"(?i)tentative[s]?\s+de\s+sui?cide",
        r"(?i)tent[ée]\s+de\s+((se\s+(suicider|tuer))|(mettre\s+fin\s+[àa]\s+((ses\s+jours?)|(sa\s+vie))))",
    ],
    "autolysis": [r"(?i)tentative\s+d'autolyse", r"(?i)autolyse"],
    "intentional_drug_overdose": [
        r"(?i)(intoxication|ingestion)\s+m[ée]dicamenteuse\s+volontaire",
        r"(?i)\b(i\.?m\.?v\.?)\b",
        r"(?i)(intoxication|ingestion)\s*([a-zA-Z0-9_éàèôê\-]+\s*){0,3}\s*volontaire",
        r"TS\s+med\s+polymedicamenteuse",
        r"TS\s+(poly)?([\s-])?m[ée]dicamenteuse",
    ],
    "jumping_from_height": [
        r"(?i)tentative[s]?\s+de\s+d[ée]fenestration",
        r"(?i)(?<!id[ée]es?\sde\s)d[ée]fenestration(?!\saccidentelle)",
        r"(?i)d[ée]fenestration\s+volontaire",
        r"(?i)d[ée]fenestration\s+intentionnelle",
        r"(?i)jet.r?\sd.un\spont",
    ],
    "cuts": [r"(?i)phl[ée]botomie"],
    "strangling": [r"(?i)pendaison"],
    "self_destructive_behavior": [r"(?i)autodestruction"],
    "burn_gas_caustic": [r"(?i)ing[eé]stion\sde\s(produit\s)?caustique"],
}



import csv
import json
import os
import warnings
from typing import Any, Dict, Optional, Union

import accelerate.tracking
from rich_logger import RichTablePrinter

import edsnlp


def flatten_dict(d, path=""):
    if not isinstance(d, (list, dict)):
        return {path: d}

    if isinstance(d, list):
        items = enumerate(d)
    else:
        items = d.items()

    return {
        k: v
        for key, val in items
        for k, v in flatten_dict(val, f"{path}/{key}" if path else key).items()
    }


@edsnlp.registry.loggers.register("csv")
class CSVLogger(accelerate.tracking.GeneralTracker):
    name = "csv"
    requires_logging_directory = True

    def __init__(
        self,
        *,
        logging_dir: Union[str, os.PathLike],
        file_name: str = "metrics.csv",
        **kwargs,
    ):
        """
        A simple CSV-based logger that writes logs to a CSV file. By default,
        with `edsnlp.train` the CSV file is located under a local directory
        `${CWD}/artifact/metrics.csv`.

        !!! warning "Consistent Keys"

            This logger expects that the `values` dictionary passed to `log` has
            consistent keys across all calls. If a new key is encountered in a
            subsequent call, it will be ignored and a warning will be issued.

        Parameters
        ----------
        logging_dir : str or os.PathLike
          Directory in which to store the CSV.
        file_name : str, optional
          Name of the CSV file. Defaults to "metrics.csv".
        """
        super().__init__()
        self.logging_dir = logging_dir

        self.file_path = os.path.join(self.logging_dir, file_name)
        self._file = None
        self._writer = None
        self._columns = None
        self._has_header = False

    @accelerate.tracking.on_main_process
    def ensure_init(self):
        if self._file is not None:
            return
        os.makedirs(self.logging_dir, exist_ok=True)
        self._file = open(self.file_path, mode="a", newline="")
        self._writer = csv.writer(self._file)

    @property
    def tracker(self):  # pragma: no cover
        return None

    @accelerate.tracking.on_main_process
    def store_init_configuration(self, values: Dict[str, Any]):
        pass

    @accelerate.tracking.on_main_process
    def log(self, values: Dict[str, Any], step: Optional[int] = None):
        """
        Logs `values` to the CSV file, at an optional `step`.

        - If it's the first call, the columns are inferred from the keys in `values`
          plus a "step" column if the user provides `step`.
        - All subsequent calls must use the same columns. Any missing columns get
          written as empty, any new columns generate a warning.
        """
        self.ensure_init()

        values = flatten_dict(values)

        if self._columns is None:
            self._columns = list({**{"step": None}, **values}.keys())
            self._writer.writerow(self._columns)
            self._has_header = True

        # Build a row in the order of self._columns
        row = []
        for col in self._columns:
            if col == "step":
                row.append(step if step is not None else "")
            else:
                if col not in values and col != "step":
                    row.append("")
                else:
                    row.append(values.get(col, ""))

        for extra_key in values.keys():
            if extra_key not in self._columns:
                warnings.warn(
                    f"CSVTracker: encountered a new field '{extra_key}' that was not in"
                    f"the field keys of the first logged step. It will not be logged."
                )

        self._writer.writerow(row)
        self._file.flush()

    @accelerate.tracking.on_main_process
    def finish(self):
        self._file.close()


@edsnlp.registry.loggers.register("json")
class JSONLogger(accelerate.tracking.GeneralTracker):
    name = "json"
    requires_logging_directory = True

    def __init__(
        self,
        logging_dir: Union[str, os.PathLike],
        file_name: str = "metrics.json",
        **kwargs,
    ):
        """
        A simple JSON-based logger that writes logs to a JSON file as a
        list of dictionaries. By default, with `edsnlp.train` the JSON file
        is located under a local directory `${CWD}/artifact/metrics.json`.

        This method is not recommended for large and frequent logging, as it
        re-writes the entire JSON file on every call. Prefer
        [`CSVLogger`][edsnlp.training.loggers.CSVLogger] for frequent
        and heavy logging.

        Parameters
        ----------
        logging_dir : str or os.PathLike
            Directory in which to store the JSON file.
        file_name : str, optional
            Name of the JSON file. Defaults to "metrics.json".
        """
        super().__init__()
        self.logging_dir = logging_dir
        self.initialized = False

        self._file_path = os.path.join(self.logging_dir, file_name)
        self._logs = []

    @property
    def tracker(self):  # pragma: no cover
        return None

    @accelerate.tracking.on_main_process
    def store_init_configuration(self, values: Dict[str, Any]):
        pass

    @accelerate.tracking.on_main_process
    def log(self, values: Dict[str, Any], step: Optional[int] = None):
        """
        Logs `values` along with a `step` (if provided).

        On every call, we:
          1. Append a new record to our in-memory list.
          2. Write out the entire JSON file containing all records.
        """
        log_entry = {"step": step, **values}
        self._logs.append(log_entry)
        os.makedirs(self.logging_dir, exist_ok=True)

        with open(self._file_path, mode="w") as f:
            json.dump(self._logs, f, indent=2)

    @accelerate.tracking.on_main_process
    def finish(self):
        pass


@edsnlp.registry.loggers.register("rich")
class RichLogger(accelerate.tracking.GeneralTracker):
    DEFAULT_FIELDS = {
        "step": {},
        "per_pipe/(.*)/results/loss": {
            "goal": "lower_is_better",
            "format": "{:.2e}",
            "goal_wait": 2,
            "name": r"\1_loss",
        },
        "lr": {"format": "{:.2e}"},
        "validation/speed/(.*)": {"format": "{:.2f}", r"name": r"\1"},
        "validation/(.*?)/micro/(f|r|p)$": {
            "goal": "higher_is_better",
            "format": "{:.2%}",
            "goal_wait": 1,
            "name": r"\1_\2",
        },
        "validation/(.*?)/(uas|las)": {
            "goal": "higher_is_better",
            "format": "{:.2%}",
            "goal_wait": 1,
            "name": r"\1_\2",
        },
        "weights/grad_norm/__all__": {
            "format": "{:.2e}",
            "name": "grad_norm",
        },
    }

    name = "rich"
    requires_logging_directory = False

    def __init__(
        self,
        run_name: Optional[str] = None,
        fields: Dict[str, Union[Dict, bool]] = None,
        key: Optional[str] = None,
        hijack_tqdm: bool = True,
        **kwargs,
    ):
        """
        A logger that displays logs in a Rich-based table using
        [rich-logger](https://github.com/percevalw/rich-logger).
        This logger is also available via the loggers registry as `rich`.

        !!! warning "No Disk Logging"

            This logger doesn't save logs to disk. It's meant for displaying
            logs in a pretty table during training. If you need to save logs
            to disk, consider combining this logger with any other logger.

        Parameters
        ----------
        fields: Dict[str, Union[Dict, bool]]
            Field descriptors containing goal ("lower_is_better" or "higher_is_better"),
             format and display name
            The key is a regex that will be used to match the fields to log
            Each entry of the dictionary should match the following scheme:

            - key: a regex to match columns
            - value: either a Dict or False to hide the column, the dict format is
                - name: the name of the column
                - goal: "lower_is_better" or "higher_is_better"

            This defaults to a set of metrics and stats that are commonly
            logged during EDS-NLP training.
        key: Optional[str]
            Key to group the logs
        hijack_tqdm: bool
            Whether to replace the tqdm progress bar with a rich progress bar.
            Indeed, rich progress bars integrate better with the rich table.
        """
        super().__init__()

        self.run_name = run_name
        fields = fields if fields is not None else self.DEFAULT_FIELDS
        self.fields = fields or {}

        self.printer = RichTablePrinter(key=key, fields=self.fields)

        if hijack_tqdm:
            self.printer.hijack_tqdm()

    @property
    def tracker(self):
        return self.printer

    @accelerate.tracking.on_main_process
    def store_init_configuration(self, values: Dict[str, Any]):
        pass

    @accelerate.tracking.on_main_process
    def log(self, values: Dict[str, Any], step: Optional[int] = None):
        """
        Logs values in the Rich table. If `step` is provided, we include it in the
        logged data.
        """
        combined = {"step": step, **flatten_dict(values)}
        self.printer.log_metrics(combined)

    @accelerate.tracking.on_main_process
    def finish(self):
        """
        Finalize the table (e.g., stop rendering).
        """
        self.printer.finalize()


@edsnlp.registry.loggers.register("tensorboard")
class TensorBoardLogger(accelerate.tracking.TensorBoardTracker):
    def __init__(
        self,
        project_name: str,
        logging_dir: Optional[Union[str, os.PathLike]] = None,
    ):
        """
        Logger for [TensorBoard](https://github.com/tensorflow/tensorboard).
        This logger is also available via the loggers registry as `tensorboard`.

        Parameters
        ----------
        project_name: str
            Name of the project.
        logging_dir: Union[str, os.PathLike]
            Directory in which to store the TensorBoard logs. Logs of different runs
            will be stored in `logging_dir/project_name`.
            The environment variable `TENSORBOARD_LOGGING_DIR` takes precedence over
            this argument.
        kwargs: Dict
            Additional keyword arguments to pass to `tensorboard.SummaryWriter`.
        """
        env_logging_dir = os.environ.get("TENSORBOARD_LOGGING_DIR", None)
        if env_logging_dir is not None and logging_dir is not None:  # pragma: no cover
            warnings.warn(
                f"Using the env TENSORBOARD_LOGGING_DIR={env_logging_dir} as the "
                f"logging directory for TensorBoard, instead of ${logging_dir}."
            )
            logging_dir = env_logging_dir
        assert logging_dir is not None, (
            "Please provide a logging directory or set TENSORBOARD_LOGGING_DIR"
        )
        super().__init__(project_name, logging_dir)

    def store_init_configuration(self, values: Dict[str, Any]):
        values = flatten_dict(values)
        return super().store_init_configuration(values)

    def log(self, values: dict, step: Optional[int] = None, **kwargs):
        values = flatten_dict(values)
        return super().log(values, step, **kwargs)


@edsnlp.registry.loggers.register("aim")
class AimLogger(accelerate.tracking.AimTracker):
    main_process_only = True

    def __init__(
        self,
        project_name: str,
        logging_dir: Optional[Union[str, os.PathLike]] = None,
        **kwargs,
    ):
        """
        Logger for [Aim](https://github.com/aimhubio/aim).

        Parameters
        ----------
        project_name: str
            Name of the project.
        logging_dir: Optional[Union[str, os.PathLike]]
            Directory in which to store the Aim logs.
            The environment variable `AIM_LOGGING_DIR` takes precedence over this
            argument.
        kwargs: Dict
            Additional keyword arguments to pass to the Aim init function.
        """
        env_logging_dir = os.environ.get("AIM_LOGGING_DIR", None)
        if env_logging_dir is not None and logging_dir is not None:
            warnings.warn(
                f"Using the env AIM_LOGGING_DIR={env_logging_dir} as the logging "
                f"directory for Aim, instead of ${logging_dir}."
            )
            logging_dir = env_logging_dir
        assert logging_dir is not None, (
            "Please provide a logging directory or set AIM_LOGGING_DIR"
        )

        from aim import Run

        self.writer = Run(
            repo=logging_dir,
            experiment=project_name,
            **kwargs,
        )
        accelerate.tracking.logger.debug(
            f"Initialized Aim run {self.writer.hash} in project {project_name}"
        )

    def log(self, values: dict, step: Optional[int], **kwargs):
        values = flatten_dict(values)
        return super().log(values, step, **kwargs)


@edsnlp.registry.loggers.register("wandb")
def WandBLogger(
    project_name: str,
    **kwargs,
) -> "accelerate.tracking.WandBTracker":  # pragma: no cover
    """
    Logger for [Weights & Biases](https://docs.wandb.ai/quickstart/).
    This logger is also available via the loggers registry as `wandb`.

    Parameters
    ----------
    project_name: str
        Name of the project. This will become the `project`
        parameter in `wandb.init`.
    kwargs: Dict
        Additional keyword arguments to pass to the WandB init function.

    Returns
    -------
    accelerate.tracking.WandBTracker
    """
    return accelerate.tracking.WandBTracker(project_name, **kwargs)


@edsnlp.registry.loggers.register("mlflow")
def MLflowLogger(
    project_name: str,
    logging_dir: Optional[Union[str, os.PathLike]] = None,
    run_id: Optional[str] = None,
    tags: Optional[Union[Dict[str, Any], str]] = None,
    nested_run: Optional[bool] = False,
    run_name: Optional[str] = None,
    description: Optional[str] = None,
) -> "accelerate.tracking.MLflowTracker":  # pragma: no cover
    """
    Logger for
    [MLflow](https://mlflow.org/docs/latest/getting-started/intro-quickstart/).
    This logger is also available via the loggers registry as `mlflow`.

    Parameters
    ----------
    project_name: str
        Name of the project. This will become the mlflow experiment name.
    logging_dir: Optional[Union[str, os.PathLike]]
        Directory in which to store the MLflow logs.
    run_id: Optional[str]
        If specified, get the run with the specified UUID and log parameters and metrics
        under that run. The run’s end time is unset and its status is set to running,
        but the run’s other attributes (source_version, source_type, etc.) are not
        changed. Environment variable MLFLOW_RUN_ID has priority over this argument.
    tags: Optional[Union[Dict[str, Any], str]]
        An optional `dict` of `str` keys and values, or a `str` dump from a `dict`, to
        set as tags on the run. If a run is being resumed, these tags are set on the
        resumed run. If a new run is being created, these tags are set on the new run.
        Environment variable MLFLOW_TAGS has priority over this argument.
    nested_run: Optional[bool]
        Controls whether run is nested in parent run. True creates a nested run.
        Environment variable MLFLOW_NESTED_RUN has priority over this argument.
    run_name: Optional[str]
        Name of new run (stored as a mlflow.runName tag). Used only when `run_id` is
        unspecified.
    description: Optional[str]
        An optional string that populates the description box of the run. If a run is
        being resumed, the description is set on the resumed run. If a new run is being
        created, the description is set on the new run.

    Returns
    -------
    accelerate.tracking.MLflowTracker
    """
    return accelerate.tracking.MLflowTracker(
        project_name,
        logging_dir=logging_dir,
        run_id=run_id,
        tags=tags,
        nested_run=nested_run,
        run_name=run_name,
        description=description,
    )


@edsnlp.registry.loggers.register("cometml")
def CometMLLogger(
    project_name: str,
    **kwargs,
) -> "accelerate.tracking.CometMLTracker":  # pragma: no cover
    """
    Logger for [CometML](https://www.comet.com/docs/).
    This logger is also available via the loggers registry as `cometml`.

    Parameters
    ----------
    project_name: str
        Name of the project.
    kwargs: Dict
        Additional keyword arguments to pass to the CometML Experiment
        object.

    Returns
    -------
    accelerate.tracking.CometMLTracker
    """
    return accelerate.tracking.CometMLTracker(project_name, **kwargs)


try:
    from accelerate.tracking import ClearMLTracker as _ClearMLTracker

    @edsnlp.registry.loggers.register("clearml")
    def ClearMLLogger(
        project_name: str,
        **kwargs,
    ) -> "accelerate.tracking.ClearMLTracker":  # pragma: no cover
        """
        Logger for
        [ClearML](https://clear.ml/docs/latest/docs/getting_started/ds/ds_first_steps/).
        This logger is also available via the loggers registry as `clearml`.

        Parameters
        ----------
        project_name: str
            Name of the experiment. Environment variables `CLEARML_PROJECT` and
            `CLEARML_TASK` have priority over this argument.
        kwargs: Dict
            Additional keyword arguments to pass to the ClearML Task object.

        Returns
        -------
        accelerate.tracking.ClearMLTracker
        """
        return _ClearMLTracker(project_name, **kwargs)
except ImportError:  # pragma: no cover

    def ClearMLLogger(*args, **kwargs):
        raise ImportError("ClearMLLogger is not available.")


try:
    from accelerate.tracking import DVCLiveTracker as _DVCLiveTracker

    @edsnlp.registry.loggers.register("dvclive")
    def DVCLiveLogger(
        live: Any = None,
        **kwargs,
    ) -> "accelerate.tracking.DVCLiveTracker":  # pragma: no cover
        """
        Logger for [DVC Live](https://dvc.org/doc/dvclive).
        This logger is also available via the loggers registry as `dvclive`.

        Parameters
        ----------
        live: dvclive.Live
            An instance of `dvclive.Live` to use for logging.
        kwargs: Dict
            Additional keyword arguments to pass to the `dvclive.Live` constructor.

        Returns
        -------
        accelerate.tracking.DVCLiveTracker
        """
        return _DVCLiveTracker(None, live=live, **kwargs)
except ImportError:  # pragma: no cover

    def DVCLiveLogger(*args, **kwargs):
        raise ImportError("DVCLiveLogger is not available.")



import pytest
from spacy.tokens import Span

import edsnlp
import edsnlp.pipes as eds

pytest.importorskip("torch.nn")


def test_ner():
    nlp = edsnlp.blank("eds")
    nlp.add_pipe(
        eds.extractive_qa(
            embedding=eds.transformer(
                model="prajjwal1/bert-tiny",
                window=20,
                stride=10,
            ),
            # During training, where do we get the gold entities from ?
            target_span_getter=["ner-gold"],
            # Which prompts for each label ?
            questions={
                "PERSON": "Quels sont les personnages ?",
                "GIFT": "Quels sont les cadeaux ?",
            },
            questions_attribute="question",
            # During prediction, where do we set the predicted entities ?
            span_setter="ents",
        ),
    )

    doc = nlp(
        "L'aîné eut le Moulin, le second eut l'âne, et le plus jeune n'eut que le Chat."
    )
    doc._.question = {
        "FAVORITE": ["Qui a eu de l'argent ?"],
    }
    # doc[0:2], doc[4:5], doc[6:8], doc[9:11], doc[13:16], doc[20:21]
    doc.spans["ner-gold"] = [
        Span(doc, 0, 2, "PERSON"),  # L'aîné
        Span(doc, 4, 5, "GIFT"),  # Moulin
        Span(doc, 6, 8, "PERSON"),  # le second
        Span(doc, 9, 11, "GIFT"),  # l'âne
        Span(doc, 13, 16, "PERSON"),  # le plus jeune
        Span(doc, 20, 21, "GIFT"),  # Chat
    ]
    nlp.post_init([doc])

    ner = nlp.pipes.extractive_qa
    batch = ner.prepare_batch([doc], supervision=True)
    results = ner.module_forward(batch)

    list(ner.pipe([doc]))[0]

    assert results["loss"] is not None
    trf_inputs = [
        seq.replace(" [PAD]", "")
        for seq in ner.embedding.tokenizer.batch_decode(batch["embedding"]["input_ids"])
    ]
    assert trf_inputs == [
        "[CLS] quels sont les cadeaux? [SEP] l'aine eut le moulin, le second eut l'ane, et [SEP]",  # noqa: E501
        "[CLS] quels sont les cadeaux? [SEP] le second eut l'ane, et le plus jeune n'eut que le [SEP]",  # noqa: E501
        "[CLS] quels sont les cadeaux? [SEP] le plus jeune n'eut que le chat. [SEP]",  # noqa: E501
        "[CLS] quels sont les personnages? [SEP] l'aine eut le moulin, le second eut l'ane, et [SEP]",  # noqa: E501
        "[CLS] quels sont les personnages? [SEP] le second eut l'ane, et le plus jeune n'eut que le [SEP]",  # noqa: E501
        "[CLS] quels sont les personnages? [SEP] le plus jeune n'eut que le chat. [SEP]",  # noqa: E501
        "[CLS] qui a eu de l'argent? [SEP] l'aine eut le moulin, le second eut l'ane, et [SEP]",  # noqa: E501
        "[CLS] qui a eu de l'argent? [SEP] le second eut l'ane, et le plus jeune n'eut que le [SEP]",  # noqa: E501
        "[CLS] qui a eu de l'argent? [SEP] le plus jeune n'eut que le chat. [SEP]",  # noqa: E501
    ]
    assert batch["targets"].squeeze(2).tolist() == [
        [0, 0, 0, 0, 4, 0, 0, 0, 0, 2, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 4, 0],
        [2, 3, 0, 0, 0, 0, 2, 3, 0, 0, 0, 0, 0, 2, 1, 3, 0, 0, 0, 0, 0, 0],
        [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
    ]

    assert nlp.config.to_yaml_str() == (
        "nlp:\n"
        "    lang: eds\n"
        "    pipeline:\n"
        "    - extractive_qa\n"
        "    tokenizer:\n"
        "        '@tokenizers': eds.tokenizer\n"
        "components:\n"
        "    extractive_qa:\n"
        "        '@factory': eds.extractive_qa\n"
        "        embedding:\n"
        "            '@factory': eds.transformer\n"
        "            model: prajjwal1/bert-tiny\n"
        "            window: 20\n"
        "            stride: 10\n"
        "        questions:\n"
        "            PERSON: Quels sont les personnages ?\n"
        "            GIFT: Quels sont les cadeaux ?\n"
        "        questions_attribute: question\n"
        "        target_span_getter:\n"
        "        - ner-gold\n"
        "        span_setter:\n"
        "            ents: true\n"
        "        infer_span_setter: false\n"
        "        mode: joint\n"
        "        window: 40\n"
        "        stride: 20\n"
    )



from dataclasses import dataclass
from typing import List, Optional, Set, Union

from spacy.tokens import Doc, Span, Token

from edsnlp.core import PipelineProtocol
from edsnlp.pipes.qualifiers.base import (
    BaseEntQualifierResults,
    BaseQualifierResults,
    BaseTokenQualifierResults,
    RuleBasedQualifier,
)
from edsnlp.pipes.terminations import termination as default_termination
from edsnlp.utils.deprecation import deprecated_getter_factory
from edsnlp.utils.filter import consume_spans, filter_spans
from edsnlp.utils.inclusion import check_inclusion
from edsnlp.utils.resources import get_verbs
from edsnlp.utils.span_getters import SpanGetterArg, get_spans

from . import patterns


def negation_getter(token: Union[Token, Span]) -> Optional[str]:
    if token._.negation is True:
        return "NEG"
    elif token._.negation is False:
        return "AFF"
    else:
        return None


@dataclass
class TokenNegationResults(BaseTokenQualifierResults):
    # Single token
    negation: bool


@dataclass
class EntNegationResults(BaseEntQualifierResults):
    # Single entity
    negation: bool
    cues: List[Span]


@dataclass
class NegationResults(BaseQualifierResults):
    # All qualified tokens and entities
    tokens: List[TokenNegationResults]
    ents: List[EntNegationResults]


class NegationQualifier(RuleBasedQualifier):
    """
    The `eds.negation` component uses a simple rule-based algorithm to detect negated
    spans. It was designed at AP-HP's EDS, following the insights of the NegEx algorithm
    by [@chapman_simple_2001].

    The component looks for five kinds of expressions in the text :

    - preceding negations, i.e., cues that precede a negated expression
    - following negations, i.e., cues that follow a negated expression
    - pseudo negations : contain a negation cue, but are not negations
      (eg "pas de doute"/"no doubt")
    - negation verbs, i.e., verbs that indicate a negation
    - terminations, i.e., words that delimit propositions.
      The negation spans from the preceding cue to the termination.

    Examples
    --------
    The following snippet matches a simple terminology, and checks the polarity of the
    extracted entities. It is complete and can be run _as is_.

    ```python
    import edsnlp, edsnlp.pipes as eds

    nlp = edsnlp.blank("eds")
    nlp.add_pipe(eds.sentences())
    # Dummy matcher
    nlp.add_pipe(eds.matcher(terms=dict(patient="patient", fracture="fracture")))
    nlp.add_pipe(eds.negation())

    text = (
        "Le patient est admis le 23 août 2021 pour une douleur au bras. "
        "Le scanner ne détecte aucune fracture."
    )

    doc = nlp(text)

    doc.ents
    # Out: (patient, fracture)

    doc.ents[0]._.negation  # (1)
    # Out: False

    doc.ents[1]._.negation
    # Out: True
    ```

    1. The result of the component is kept in the `negation` custom extension.

    Extensions
    ----------
    The `eds.negation` component declares two extensions, on both `Span` and `Token`
    objects :

    1. The `negation` attribute is a boolean, set to `True` if the component predicts
       that the span/token is negated.
    2. The `negation_` property is a human-readable string, computed from the `negation`
       attribute. It implements a simple getter function that outputs `AFF` or `NEG`,
       depending on the value of `negation`.

    Performance
    -----------
    The component's performance is measured on three datasets :

    - The ESSAI ([@dalloux2017ESSAI]) and CAS ([@grabar2018CAS]) datasets were developed
      at the CNRS. The two are concatenated.
    - The NegParHyp corpus was specifically developed at AP-HP to test the component
      on actual clinical notes, using pseudonymised notes from the AP-HP.

    | Dataset   | Negation F1 |
    |-----------|-------------|
    | CAS/ESSAI | 71%         |
    | NegParHyp | 88%         |

    !!! note "NegParHyp corpus"

        The NegParHyp corpus was built by matching a subset of the MeSH terminology with
        around 300 documents from AP-HP's clinical data warehouse. Matched entities were
        then labelled for negation, speculation and family context.

    Parameters
    ----------
    nlp : PipelineProtocol
        The pipeline object.
    name : Optional[str]
        The component name.
    attr : str
        spaCy's attribute to use
    pseudo : Optional[List[str]]
        List of pseudo negation cues.
    preceding : Optional[List[str]]
        List of preceding negation cues
    preceding_regex : Optional[List[str]]
        List of preceding negation cues, but as regexes.
    following : Optional[List[str]]
        List of following negation cues.
    verbs : Optional[List[str]]
        List of negation verbs.
    termination : Optional[List[str]]
        List of termination terms.
    span_getter : SpanGetterArg
        Which entities should be classified. By default, `doc.ents`
    on_ents_only : Union[bool, str, List[str], Set[str]]
        Deprecated, use `span_getter` instead.

        Whether to look for matches around detected entities only.
        Useful for faster inference in downstream tasks.

        - If True, will look in all ents located in `doc.ents` only
        - If an iterable of string is passed, will additionally look in `doc.spans[key]`
        for each key in the iterable
    within_ents : bool
        Whether to consider cues within entities.
    explain : bool
        Whether to keep track of cues for each entity.

    Authors and citation
    --------------------
    The `eds.negation` component was developed by AP-HP's Data Science team.
    """

    default_patterns = patterns

    def __init__(
        self,
        nlp: PipelineProtocol,
        name: Optional[str] = "negation",
        *,
        pseudo: Optional[List[str]] = None,
        preceding: Optional[List[str]] = None,
        preceding_regex: Optional[List[str]] = None,
        following: Optional[List[str]] = None,
        verbs: Optional[List[str]] = None,
        termination: Optional[List[str]] = None,
        attr: str = "NORM",
        span_getter: Optional[SpanGetterArg] = None,
        on_ents_only: Optional[Union[bool, str, List[str], Set[str]]] = None,
        within_ents: bool = False,
        explain: bool = False,
    ):
        terms = dict(
            pseudo=patterns.pseudo if pseudo is None else pseudo,
            preceding=patterns.preceding if preceding is None else preceding,
            following=patterns.following if following is None else following,
            termination=default_termination if termination is None else termination,
            verbs=patterns.verbs if verbs is None else verbs,
        )
        terms["verbs_preceding"], terms["verbs_following"] = self.load_verbs(
            terms["verbs"]
        )
        regex = dict(
            preceding=preceding_regex
            if preceding_regex is not None
            else patterns.preceding_regex,
        )

        super().__init__(
            nlp=nlp,
            name=name,
            attr=attr,
            explain=explain,
            terms=terms,
            regex=regex,
            attributes=["_.negation"],
            on_ents_only=on_ents_only,
            span_getter=span_getter,
        )

        self.within_ents = within_ents
        self.set_extensions()

    def set_extensions(self) -> None:
        super().set_extensions()
        for cls in (Token, Span):
            if not cls.has_extension("negation"):
                cls.set_extension("negation", default=None)

            if not cls.has_extension("negated"):
                cls.set_extension(
                    "negated", getter=deprecated_getter_factory("negated", "negation")
                )

            if not cls.has_extension("negation_"):
                cls.set_extension(
                    "negation_",
                    getter=negation_getter,
                )

            if not cls.has_extension("polarity_"):
                cls.set_extension(
                    "polarity_",
                    getter=deprecated_getter_factory("polarity_", "negation_"),
                )

        if not Span.has_extension("negation_cues"):
            Span.set_extension("negation_cues", default=[])

    def load_verbs(self, verbs: List[str]) -> List[str]:
        """
        Conjugate negating verbs to specific tenses.

        Parameters
        ----------
        verbs: list of negating verbs to conjugate

        Returns
        -------
        list_neg_verbs_preceding: List of conjugated negating verbs preceding entities.
        list_neg_verbs_following: List of conjugated negating verbs following entities.
        """

        neg_verbs = get_verbs(verbs)

        neg_verbs_preceding = neg_verbs.loc[
            ((neg_verbs["mode"] == "Indicatif") & (neg_verbs["tense"] == "Présent"))
            | (neg_verbs["tense"] == "Participe Présent")
            | (neg_verbs["tense"] == "Participe Passé")
            | (neg_verbs["tense"] == "Infinitif Présent")
        ]
        neg_verbs_following = neg_verbs.loc[neg_verbs["tense"] == "Participe Passé"]
        list_neg_verbs_preceding = list(neg_verbs_preceding["term"].unique())
        list_neg_verbs_following = list(neg_verbs_following["term"].unique())

        return list_neg_verbs_preceding, list_neg_verbs_following

    def __call__(self, doc: Doc) -> Doc:
        results = self.process(doc)
        if not self.on_ents_only:
            for token_results in results.tokens:
                token_results.token._.negation = (
                    token_results.token._.negation or token_results.negation
                )
        for ent_results in results.ents:
            ent, cues, negation = (
                ent_results.ent,
                ent_results.cues,
                ent_results.negation,
            )
            ent._.negation = ent._.negation or negation

            if self.explain and negation:
                ent._.negation_cues += cues

            if not self.on_ents_only and negation:
                for token in ent:
                    token._.negation = True
        return doc

    def process(self, doc_like: Union[Doc, Span]) -> NegationResults:
        doc = doc_like.doc if isinstance(doc_like, Span) else doc_like
        entities = list(get_spans(doc_like, self.span_getter))
        matches = self.get_cues(doc_like, entities if self.on_ents_only else None)

        terminations = [m for m in matches if m.label_ == "termination"]
        boundaries = self._boundaries(doc_like, terminations)

        # Removes duplicate matches and pseudo-expressions in one statement
        matches = filter_spans(matches, label_to_remove="pseudo")

        ents = None

        token_results, ent_results = [], []

        for start, end in boundaries:
            ents, entities = consume_spans(
                entities,
                filter=lambda s: check_inclusion(s, start, end),
                second_chance=ents,
            )

            sub_matches, matches = consume_spans(
                matches, lambda s: start <= s.start < end
            )

            if self.on_ents_only and not ents:
                continue

            sub_preceding = [m for m in sub_matches if m.label_ == "preceding"]
            sub_following = [m for m in sub_matches if m.label_ == "following"]
            # Verbs preceding negated content
            sub_preceding += [m for m in sub_matches if m.label_ == "verbs_preceding"]
            # Verbs following negated content
            sub_following += [m for m in sub_matches if m.label_ == "verbs_following"]

            if not self.on_ents_only:
                for token in doc[start:end]:
                    token_results.append(
                        TokenNegationResults(
                            token=token,
                            negation=(
                                any(m.end <= token.i for m in sub_preceding)
                                or any(m.start > token.i for m in sub_following)
                            ),
                        )
                    )

            for ent in ents:
                if self.within_ents:
                    cues = [m for m in sub_preceding if m.end <= ent.end]
                    cues += [m for m in sub_following if m.start >= ent.start]
                else:
                    cues = [m for m in sub_preceding if m.end <= ent.start]
                    cues += [m for m in sub_following if m.start >= ent.end]

                negation = bool(cues)

                ent_results.append(
                    EntNegationResults(
                        ent=ent,
                        cues=cues,
                        negation=negation,
                    )
                )

        return NegationResults(tokens=token_results, ents=ent_results)
